/*\
|*| -------------------------------------------------------------------------
|*|     Cirrus Logic 542x chipset driver
|*| -------------------------------------------------------------------------
|*|
|*|     Copyright (C) 1996       Iain Riley,  [iain@iinet.net.au]
|*|                              Martin Erhardsen, [u930534@daimi.aau.dk]
|*|
\*/

#define	MAINTAINER	"Martin Erhardsen, [u930534@daimi.aau.dk]"

#undef DEBUG_REGS

extern void	request_region	( unsigned int from, unsigned int num,
				  const char *name ) ;
extern int	check_region	( unsigned int from, unsigned int num ) ;
extern void	release_region	( unsigned int from, unsigned int num ) ;

extern struct kgi_clock clock;
extern struct kgi_ramdac ramdac;

#define min(a,b) (((a)<(b))?(a):(b)) 

/*\
|*|	GGI Includes
\*/

#define	DRIVER_VERSION	"v0.05"

#include <ggi/graphdev.h>

#include <ggi/Cirrus/542x.h>
#include "chip_cmd.inc"

static struct kgi_chipset
chipinfo =
{
	"Cirrus Logic",		/* manufacturer		*/
	"542x",			/* model		*/
	GM_TEXT |               /* supported modes	*/ 

	/* Comment out this line to disable planar modes */
        GM_1BIT | GM_4BIT | GM_4BIT_PLANES | 

	GM_8BIT | GM_15BIT | GM_16BIT | GM_24BIT,
	0,
	{ 2087, 1025 },		/* maximum resolution	*/
	0,			/* ram size */
	{ 230934, 86000000 }	/* DCLK limits */
} ;

static struct statetable
default_state =
{
	0x2F,							/* MISC */
	0x00,							/* FCTRL */
        0x00,                                                   /* HDR */

	{							/* SEQ */
	0x03, 0x00, 0x0F, 0x00, 0x02, 0x0F, 0x5B, 0x00, 
	0x80, 0x0C, 0x11, 0x4A, 0x5B, 0x42, 0x2F, 0x14, 

	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xD8, 0x31,         /* 0x1x */ 
	0x00, 0x01, 0x00, 0x2B, 0x2F, 0x1F, 0x2A, 0x1C 

	},

	{							/* CRT */
	0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,
	0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

	0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE3,		/* 0x1x */
	0xFF, 0x00, 0x02, 0x02
	},

	{							/* GRC */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0F,
	0xFF, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00 
	},

	{							/* ATC */
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,

	0x00, 0x00, 0x0F, 0x00, 0x00				/* 0x1x */
	},
} ;

static int chip_version ;
static unsigned long graph_buffer = 0;  /*  Disable linear adressing */
/*static unsigned long graph_buffer = 0x000E00000; */
/*   Address of linear mapping */


static unsigned char curr_page = 0; /* Currently mapped bank */
#define SET_PAGE(bank) outbGRC((curr_page=(bank))<<((chip_version<CLGD5426_ID)?4:2),0x09)
#define GET_PAGE curr_page



static char *
chip_strings [ 10 ] =
	{
	"5420",		/* CLDG5420 */
	"5422",		/* CLDG5422 */
	"5426",		/* CLDG5426 */
	"5424",		/* CLDG5424 */
	"5428",		/* CLDG5428 */
	"5429",		/* CLDG5429 */
	"5430",		/* CLDG5430 */
	"",
	"5434",		/* CLDG5434 */
	"5436",		/* CLDG5436 */
	} ;

static char *
bus_strings [ 4 ] =
	{
	"UNKNOWN",
	"ISA",
	"VLB",
	"PCI"
	} ;

#ifdef DEBUG_REGS

void
printk_registers ( void )
{
int	temp, x ;

	printk ( "MISC: %X\n", inbMISC ( ) ) ;

	printk ( "\nSEQ:" ) ;
	for ( x = 0 ; x < NoSEQRegs ; x++ )
		{
		if ( ! ( x & 0x0F ) )
			printk ( "\n%.2X: ", x ) ;
		printk ( "%.2X, ", inbSEQ ( x ) ) ;
		}

	printk ( "\n\nCRT:" ) ;
	for ( x = 0 ; x < NoCRTRegs ; x++ )
		{
		if ( ! ( x & 0x0F ) )
			printk ( "\n%.2X: ", x ) ;
		printk ( "%.2X, ", inbCRT ( x ) ) ;
		}

	printk ( "\n\nGRC:" ) ;
	for ( x = 0 ; x < NoGRCRegs ; x++ )
		{
		if ( ! ( x & 0x0F ) )
			printk ( "\n%.2X: ", x ) ;
		printk ( "%.2X, ", inbGRC ( x ) ) ;
		}

	printk ( "\n\nATC:" ) ;
	for ( x = 0 ; x < NoATCRegs ; x++ )
		{
		if ( ! ( x & 0x0F ) )
			printk ( "\n%.2X: ", x ) ;
		printk ( "%.2X, ", inbATC ( x ) ) ;
		}

	printk ( "\n" ) ;
}

#endif


static void init_542x(void)
{  
  int maxVCLK;
  int DRAMbw;
  /* Detect MCLK */
  unsigned char MCLK;
  static const unsigned char fixedMCLK[4] = {0x1c, 0x19, 0x17, 0x15};
  if (chip_version>=CLGD5424_ID){ 
    MCLK=inbSEQ(0x1F)&0x3F;
  } else {
    MCLK=fixedMCLK[inbSEQ(0x0F)&3];
  };
  default_state.SEQ[0x1F]=MCLK;

  DRAMbw = ( 14318000 * (int)MCLK )>>4;
  if (chipinfo.ramsize >= (1<<19)) DRAMbw *=2;
  if (chipinfo.ramsize >= (1<<20)) DRAMbw *=2;
  if ((chipinfo.ramsize >= (1<<21))
      && (chip_version>=CLGD5434_ID)) DRAMbw *=2;
  /* Leave some bandwidth */
  DRAMbw = (DRAMbw * 10)/11;

  maxVCLK = (chip_version>=CLGD5428_ID) ? 86000000 : 80000000;
  chipinfo.dclk.max=maxVCLK;
  ramdac.normal.max=maxVCLK;
  clock.clock.range[0].max=maxVCLK;

  /* Init ramdac structure */
  ramdac.clock[GT_1BIT].max=maxVCLK;
  ramdac.clock[GT_4BIT_PLANES].max=maxVCLK;
  ramdac.clock[GT_8BIT].max=min(maxVCLK,DRAMbw);
  ramdac.clock[GT_15BIT].max=
    ramdac.clock[GT_15BIT].max= min(
    (chip_version>=CLGD5426_ID) ? maxVCLK : maxVCLK/2,DRAMbw/2);
  ramdac.clock[GT_24BIT].max=min(maxVCLK/3,DRAMbw/3);
  if (chip_version>=CLGD5434_ID)
    {
      ramdac.clock[GT_32BIT].max=min(maxVCLK,DRAMbw/4);
      chipinfo.modes |= GT_32BIT;
    }
  ramdac.modes=chipinfo.modes;
};  

static int
detect_542x ( void )
{
int	lock ;
int	memory_size ;
int	cirrus_bus ;

	lock = inbSEQ ( 0x06 ) ;

	if ( lock != 0x12 )		/* First Check for Cirrus Chipset */
		return -ECHIPSET ;

	chip_version = inbCRT ( 0x27 ) >> 2 ;

	switch ( chip_version )
		{
		case	CLGD5420_ID:
		case	CLGD5422_ID:
		case	CLGD5424_ID:
		case	CLGD5426_ID:
		case	CLGD5428_ID:
		case	CLGD5429_ID:

		case	CLGD5430_ID:
		case	CLGD5434_ID:
		case	CLGD5436_ID:
/*				printk ( "Cirrus Logic %s Detected\n", chip_strings [ chip_version - 0x22 ] ) ; */
				break ;

		default:
			return	-ECHIPSET ;
		}

	memory_size = 256 << ( ( inbSEQ ( 0x0A ) >> 3 ) & 0x03 ) ;
	chipinfo.ramsize = memory_size * 1024 ;

	switch ( ( inbSEQ ( 0x17 ) >> 3 ) & 0x07 )
		{
		case	2:
		case	6:	cirrus_bus = CIRRUS_BUS_VLB ;	
				break ;

		case	4:	cirrus_bus = CIRRUS_BUS_PCI ;
				break ;

		case	7:	cirrus_bus = CIRRUS_BUS_ISA ;
				break ;

		default:	cirrus_bus = CIRRUS_BUS_UNKNOWN ;
		}

	printk ( "Cirrus Logic %s, %s bus %dK Detected\n", chip_strings [ chip_version - 0x22 ], bus_strings [ cirrus_bus ], memory_size ) ;

	if (!graph_buffer) 
	  {
	    fb_graph_len = 0x10000 ;		/* map 64k at 0xA0000 */
	    fb_graphics = (void *) 0x000A0000 ;
	  }
	else
	  {
	    fb_graph_len = chipinfo.ramsize ;
	    fb_graphics = vremap(graph_buffer,fb_graph_len);
	  }

	if ( fb_graphics == NULL )
		{
		printk ( "%s: no physical address space (graphics framebuffer).\n", __FILE__ ) ;
		return -ECHIPSET ;
		}

	printk ( "text fb: 0x%p graphics fb: 0x%p\n",(void *)0xB8000, fb_graphics ) ;
	init_542x();

	return EOK ;
}

#define	SETFONT	0
#define	GETFONT	1

static void
getsetfont ( int get, int no, ggi_font *font )
{
	IS_CRITICAL ;

unsigned char	tmp,
		SR2,
		SR4,
		GR4,
		GR5,
		GR6 ;
int		i ;
char		*fbuf,
		*data ;

	if ( font == NULL )
		return ;

	fbuf = ( char* ) ( 0x000A0000 + ( no * 0x2000 ) + ( font->first << 5 ) ) ;
	data = font->data ;

	ENTER_CRITICAL ;

		SR2 = inbSEQ ( SEQ_WPLANE ) ;
		SR4 = inbSEQ ( SEQ_MEMMODE ) ;
		GR4 = inbGRC ( GRC_READPLANE ) ;
		GR5 = inbGRC ( GRC_MODECONTROL ) ;
		GR6 = inbGRC ( GRC_MMAP_MODE ) ;

		outbSEQ ( SR02_PLANE2, SEQ_WPLANE ) ;		/* Write to Plane 2 */
		outbSEQ ( SR04_256K_ACCESS | SR04_NO_ODDEVEN, SEQ_MEMMODE ) ;

		outbGRC ( 2, GRC_READPLANE ) ;			/* Read from Plane 2 */
		outbGRC ( GR5_WRITEMODE0, GRC_MODECONTROL ) ;
		outbGRC ( GR6_MAP_A0_64 | GR6_GRAPHMODE, GRC_MMAP_MODE ) ;

	LEAVE_CRITICAL ;

	if ( get == GETFONT )
		{
		if ( data )
			for ( i = font->first; i <= font->last; i++ )
				{
				memmove ( data, fbuf, font->ysize ) ;
				fbuf += 32;
				data += font->ysize ;
				}
		}
	else
		if ( data )
			{
			register int	cnt = 32 - font->ysize ;

			for ( i = font->first; i <= font->last; i++ )
				{
				memmove ( fbuf, data, font->ysize ) ;
				memset ( fbuf + font->ysize, 0, cnt ) ;
				fbuf += 32;
				data += font->ysize;
				}
			}
		else
			memset ( fbuf, 0, ( font->last - font->first + 1 ) << 5 ) ;

	ENTER_CRITICAL ;

		tmp = ( inbCRT ( CRT_CURSORSTART ) & ~CR0A_CURSOR_START_MASK ) ;
		outbCRT ( tmp | ( font->cr_start & CR0A_CURSOR_START_MASK ) , CRT_CURSORSTART ) ;
		tmp = ( inbCRT ( CRT_CURSOREND ) & ~CR0B_CURSOR_END_MASK ) ;
		outbCRT ( tmp | ( font->cr_end & CR0B_CURSOR_END_MASK ) , CRT_CURSOREND ) ;

		#ifndef UNDERLINE
		font->underline = CR14_UNDERLINE_MASK ;
		#endif

		tmp = ( inbCRT ( CRT_UNDERLINE ) & ~CR14_UNDERLINE_MASK ) ;
		outbCRT ( tmp | ( font->underline & CR14_UNDERLINE_MASK ) , CRT_UNDERLINE ) ;

		outbSEQ ( SR2, SEQ_WPLANE ) ;
		outbSEQ ( SR4, SEQ_MEMMODE ) ;
		outbGRC ( GR4, GRC_READPLANE ) ;
		outbGRC ( GR5, GRC_MODECONTROL ) ;
		outbGRC ( GR6, GRC_MMAP_MODE ) ;

	LEAVE_CRITICAL ;
}

/*\
|*|	GGI Interface
\*/

int
kgi_chipset_init ( void )
{
	TRACE ( printk ( "Cirrus Logic 542x Chipset Driver "DRIVER_VERSION"\n" ) ) ;

	cl_unlock ( ) ;

	if ( ! detect_542x ( ) )
		{
		if ( ! check_region ( CLIO_BASE, CLIO_REGION ) )
			request_region ( CLIO_BASE, CLIO_REGION, "Cirrus Logic 542x" ) ;
		else
			printk ( "%s: warning: failed to allocated IO region.\n", __FILE__ ) ;

		return EOK ;
		}
	else
		return -ECHIPSET ;
}

void
kgi_chipset_done ( void )
{
	release_region ( CLIO_BASE, CLIO_REGION ) ;

	TRACE ( printk ( "Cirrus Logic Chipset Driver removed\n" ) ) ;

	cl_lock ( ) ;
}

struct kgi_chipset
*kgi_chipset_get_info ( void )
{
	return &chipinfo ;
}

static void
adjust_timing ( struct kgi_mode *mode )
{
register int foo = ( mode->request.graphtype == GT_TEXT )
			? mode->request.textgrid.x : 8 ;

#define ADJUST(X) mode->x.X -= ( mode->x.X % foo )

	ADJUST ( width ) ;
	ADJUST ( syncstart ) ;
	ADJUST ( syncend ) ;
	ADJUST ( blankstart ) ;
	ADJUST ( blankend ) ;

	if ( mode->x.total % foo )
		mode->x.total += foo - ( mode->x.total % foo ) ;

	if ( ( ( mode->x.syncend / foo ) & 0x3F ) == ( ( mode->x.syncstart / foo ) & 0x3F ) )
		mode->x.syncend -= foo ;
	if ( ( ( mode->x.blankend / foo ) & 0x7F ) == ( ( mode->x.blankstart / foo ) & 0x7F ) )
		mode->x.blankend -= foo ;

	if ( ( mode->y.syncend & 0x0F ) == ( mode->y.syncstart & 0x0F ) )
		mode->y.syncend-- ;
	if ( ( mode->y.blankend & 0xFF ) == ( mode->y.blankstart & 0xFF ) )
		mode->y.blankend-- ;

#undef ADJUST
}

static void
set_origin ( unsigned int offset )
{
	IS_CRITICAL ;

	ENTER_CRITICAL ;

		outbCRT ( ((offset>>16)&0x01) |
                          ((offset>>17)&0x03)<<2 |
			  inbCRT(0x1B), 0x1B);
		outbCRT ( offset >> 8, CRT_STARTADDR_H ) ;
		outbCRT ( offset, CRT_STARTADDR_L ) ;

	LEAVE_CRITICAL ;
}

static void
show_cursor ( struct kgi_pointer *cursor, unsigned int offset )
{
	IS_CRITICAL ;

	ENTER_CRITICAL ;

		cursor->state |= PF_SHOWN ;
		outbCRT ( offset >> 8, CRT_CURSORADDR_H ) ;
		outbCRT ( offset, CRT_CURSORADDR_L ) ;

	LEAVE_CRITICAL ;
}

static void
hide_cursor ( struct kgi_pointer *cursor, unsigned int offset )
{
	IS_CRITICAL ;

	if ( ! ( cursor->state & PF_SHOWN ) )	/* Is cursor already off */
		return ;

	ENTER_CRITICAL ;

		cursor->state &= ~PF_SHOWN ;
		outbCRT ( 0xFF, CRT_CURSORADDR_H ) ;
		outbCRT ( 0xFF, CRT_CURSORADDR_L ) ;

	LEAVE_CRITICAL ;
}

static int sel_page = 0;

int
kgi_chipset_check_mode ( struct kgi_mode *mode, enum kgi_tm_cmd cmd )
{
  int	bpp = 1 ;
  
  switch ( mode->request.graphtype )
    {
    case GT_32BIT:
      if (chip_version<CLGD5430_ID) return -ECHIPSET;
      bpp=32; break;
    case GT_24BIT:
      bpp=24;mode->clk.mul=3;
      break;
    case  GT_16BIT:
    case  GT_15BIT:	
      if (chip_version<CLGD5426_ID) mode->clk.mul=2;
      bpp = 16; break ; 
    case  GT_8BIT:	
      bpp = 8 ; break ;
    case GT_4BIT:     TRACE(printk("%s: Fallback GT_4BIT -> GT_4BIT_PLANES\n",__FILE__));
      mode->request.graphtype=GT_4BIT_PLANES;
    case	GT_4BIT_PLANES:	
      bpp = 4 ; break ;
    case	GT_1BIT: 
      bpp = 1 ; break ;
    case	GT_TEXT:	break ;
    default:
      TRACE ( printk ( "%s: graphtype %i not supported\n", __FILE__, mode->request.graphtype ) ) ;
      return -ECHIPSET ;
    }
  
  switch ( cmd )
    {
    case CMD_PROPOSE:
      if ( mode->request.graphtype == GT_TEXT )
	{
	  mode->magnify.x = mode->magnify.y = 1 ;
	  if ( ( mode->request.textgrid.x > 9 ) || ( mode->request.textgrid.x < 7 ) )
	    {
	      printk ( "Invalid Font Size - %d\n", mode->request.textgrid.x ) ;
	      return -EINVAL ;
	    }

	  if ( ( 2 * ( mode->request.virtual.x / mode->request.textgrid.x )
		 * ( mode->request.virtual.y / mode->request.textgrid.y ) ) > 0x8000 )
	    {
	      printk ( "No Memory for screen\n" ) ;
	      return -ENOMEM ;
	    }
	  
	  mode->request.visible.x -=
	    ( mode->request.visible.x % mode->request.textgrid.x ) ;
	  mode->request.visible.y -=
	    ( mode->request.visible.y % mode->request.textgrid.y ) ;
	}
      else		/* A graphic mode */
	{
	  mode->magnify.x = 1 ;

	  /* Hardware limit */
	  mode->request.visible.x &= 0x7FFFFFF8 ;
	  
	  mode->request.visible.y *= 
	    (mode->magnify.y = (350+ mode->request.visible.y -1 )
	     / mode->request.visible.y);

	  if ( mode->magnify.y > 15 ) return -EINVAL ;

          /* Limit caused by offset resister */
          switch (mode->request.graphtype)
            {
            case GT_1BIT:
            case GT_4BIT_PLANES:
              mode->request.virtual.x += 0xF ;
              mode->request.virtual.x &= 0x7FFFFFF0 ;
            case GT_8BIT:
            case GT_24BIT:
              mode->request.virtual.x += 0x7 ;
              mode->request.virtual.x &= 0x7FFFFFF8 ;
            case GT_15BIT:
            case GT_16BIT:
              mode->request.virtual.x += 0x3 ;
              mode->request.virtual.x &= 0x7FFFFFFC ;
            case GT_32BIT:
              mode->request.virtual.x += 0x1 ;
              mode->request.virtual.x &= 0x7FFFFFFE ;
            default:
            }
	  
	  if ( ( mode->request.virtual.x * mode->request.virtual.y * bpp / 8 )
	       > chipinfo.ramsize )
	    {
	      printk ( "%s: not enough memory to do virtual %ix%i at %i bpp - Ram Size = %i\n", __FILE__, mode->request.virtual.x, mode->request.virtual.y, bpp, chipinfo.ramsize ) ;
	      return -ENOMEM ;
	    }

	}

      mode->x.width = mode->request.visible.x ;
      mode->y.width = mode->request.visible.y ;

      return EOK ;
      
    case	CMD_LOWER:
      if ( mode->dclk < chipinfo.dclk.min )
	{
	  TRACE(printk ( "%s: DCLK too low (%i).\n", __FILE__, mode->dclk ));
	  return -ECHIPSET ;
	}

      mode->dclk = ( mode->dclk > chipinfo.dclk.max )
	? chipinfo.dclk.max : mode->dclk ;
      
      adjust_timing ( mode ) ;
      
      return EOK ;

    case	CMD_RAISE:
      if ( mode->dclk > chipinfo.dclk.max )
	{
	  TRACE(printk("%s: DCLK too high (%i).\n", __FILE__, mode->dclk ));
	  return -ECHIPSET ;
	}

      mode->dclk = ( mode->dclk < chipinfo.dclk.min )
	? chipinfo.dclk.min : mode->dclk ;
      
      adjust_timing ( mode ) ;
      
      return EOK ;
      
    case	CMD_CHECK:
      if ( mode->request.graphtype == GT_TEXT )
	{
	  mode->set_origin = set_origin ;
	  mode->cursor_hide = hide_cursor ;
	  mode->cursor_show = show_cursor ;
	  mode->flags |= FL_CHIP_COLORTEXT ;
	  if ((mode->magnify.x != 1) || (mode->magnify.y != 1)	/*driver */
	      ||(mode->request.textgrid.x < 8)	/* hardware */
	      ||(9 < mode->request.textgrid.x)	/* hardware */
	      ||(mode->request.textgrid.y < 1)	/* hardware */
	      ||(mode->request.textgrid.y > 32)	/* hardware */
	      ||((2 * (mode->request.virtual.x / mode->request.textgrid.x)
		  * (mode->request.virtual.y 
		     / mode->request.textgrid.y)) > 0x8000)
	      || (mode->dclk > chipinfo.dclk.max)		/* hardware */
	      ||(mode->dclk < chipinfo.dclk.min)) /* hardware */
	    {	/* hardware */
	      TRACE(printk("%s: hw check failed\n", __FILE__));
	      return -ECHIPSET;
	    }
	}
      else if (((mode->request.virtual.x * 
		 mode->request.virtual.y * bpp / 8) > chipinfo.ramsize)
	       || (mode->magnify.x != 1) || (mode->magnify.y > 15)
	       || (mode->dclk < chipinfo.dclk.min)
	       || (chipinfo.dclk.max < mode->dclk))
	       {
		 TRACE(printk("%s: hw check failed\n", __FILE__));
		 return -ECHIPSET;
	       }

#define TM(X) (((unsigned) chipinfo.maxdots.x < mode->x.X)||((unsigned) chipinfo.maxdots.y < mode->y.X))
      if (TM(width) || TM(blankstart) || TM(syncstart)
	  || TM(syncend) || TM(blankend) || TM(total)) 
	{
	  TRACE(printk("%s: timing check failed\n", __FILE__));
	  return -ECHIPSET;
	}
#undef TM

      adjust_timing ( mode ) ;
      return EOK ;
      
    default:
      printk ( "%s: Unknown CMD received %d\n", __FILE__, cmd ); 
      return -ECHIPSET ;
    }
}

void
kgi_chipset_set_mode ( struct kgi_mode *mode )
{
  IS_CRITICAL ;
  struct statetable	setup = default_state ;
  register int		foo;
  int			DCLKdiv = 1 ;
  
  switch ( mode->request.graphtype )
    {
    case GT_TEXT: foo = 2*mode->request.virtual.x / mode->request.textgrid.x ;
      foo /= 4 ;break ;
    case GT_4BIT_PLANES:
    case GT_1BIT:   foo =  mode->request.virtual.x>>4   ; break ;
    case GT_8BIT:   foo = (mode->request.virtual.x>>3)  ; break ;
    case GT_15BIT:
    case GT_16BIT:  foo = (mode->request.virtual.x>>3)*2; break ;
    case GT_24BIT:  foo = (mode->request.virtual.x>>3)*3; break ;
    case GT_32BIT:  foo = (mode->request.virtual.x>>3)*4; break ;
    default:		INTERNALERROR ;
					return ;
    }

  setup.CRT [ CRT_LOGICALWIDTH ] |= foo & 0xFF ;
  setup.CRT [ 0x1B ] |= (foo&0x100) ? 0x08 : 0x00;
  
  if ( mode->request.virtual.y < 350 )
    {
      mode->magnify.y = mode->request.visible.y / mode->request.virtual.y ;
    }

  if ( mode->request.graphtype == GT_TEXT )
    {
      setup.CRT [ CRT_MODE ] = CR17_CGA_ADDR_WRAP | CR17_WORDMODE
	| CR17_ENABLE_SYNC | CR17_CGA_BANKING | CR17_HGC_BANKING ;
      setup.SEQ [0x02] = SR02_PLANE0 | SR02_PLANE1 ;
      setup.GRC [0x05] |= GR5_ODD_EVEN ;
      setup.GRC [0x06] |= GR6_MAP_B8_32 | GR6_CHAIN_OE ;
      
      switch ( ( DCLKdiv = mode->request.textgrid.x ) )
	{
	case 8:	setup.SEQ [0x01] |= SR01_8DOT_CHCLK ; break ;
	case 9:	break ;
	default:	INTERNALERROR ;
	  return ;
	}
      
      setup.CRT [0x09] |= mode->request.textgrid.y - 1 ;
      setup.ATC [0x13] = ( mode->request.textgrid.x == 9 ) ? 0x08 : 0x00 ;
    }
  else
    {
      DCLKdiv = 8 ;
      
      setup.GRC[0x0B] = 0x20; /* enable bank switching */
      setup.SEQ[0x07] |= (graph_buffer) ? 0xE0 : 0x00;

      setup.CRT [0x17] = CR17_CGA_BANKING | CR17_HGC_BANKING
	| CR17_ENABLE_SYNC | CR17_CGA_ADDR_WRAP | CR17_BYTE_MODE;

      for ( foo = 0 ; foo < 16 ; foo++ ) setup.ATC [ foo ] = foo ;

      setup.SEQ [0x02] |= SR02_PLANE_MASK ;
      
      setup.ATC [0x10] |= AR10_GRAPHICS ;
		
      setup.GRC [0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64 ;
      
      setup.CRT [0x09] |= ( mode->magnify.y - 1 ) & CR09_CHARHEIGHT_MASK ;

      setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
      setup.HDR = 0x00; 

      switch (mode->request.graphtype)
	{
	case GT_1BIT:
	  setup.SEQ[0x04] |= SR04_NO_ODDEVEN | SR04_256K_ACCESS;
	  setup.ATC[0x12] = 0x01; /* Use 1 bit */
	  break;
	case GT_4BIT_PLANES:
	  setup.GRC[0x01] |= 0x0f;      /* Enable all planes */
	  setup.SEQ[0x04] |= SR04_NO_ODDEVEN | SR04_256K_ACCESS;
	  setup.GRC[0x05] |= GR5_WRITEMODE3;
	  setup.GRC[0x08] |= 0xff;      /* Enable all bits */
	  break;
	case GT_8BIT:
	  setup.SEQ[0x04] |= 0x0E;
	  setup.SEQ[0x07] |= 0x01;
	  setup.GRC[0x05] |= 0x40;
	  break;
	case GT_15BIT:
	  setup.SEQ[0x04] |= 0x0E;
	  setup.SEQ[0x07] |= 0x07;
	  setup.GRC[0x05] |= 0x40;
	  setup.HDR=0xE0;
	  break;
	case GT_16BIT:
	  setup.SEQ[0x04] |= 0x0E;
	  setup.SEQ[0x07] |= 0x07;
	  setup.GRC[0x05] |= 0x40 ;
	  setup.HDR=0xE1;
	  break;
	case GT_24BIT:
	  setup.SEQ[0x04] |= 0x0E;
	  setup.SEQ[0x07] |= 0x05;
	  setup.GRC[0x05] |= 0x40 ;
	  setup.HDR=0xE5;
	  break;
	case GT_32BIT:
	  setup.SEQ[0x04] |= 0x0E;
	  setup.SEQ[0x07] |= 0x09;
	  setup.GRC[0x05] |= 0x40 ;
	  setup.HDR=0xE5;
	  break;
	default:
	};
    }

  /*\
  |*|	setup horizontal timing
  \*/
 
   foo = (mode->x.width / DCLKdiv) - 1;		/* horizontal display end */
     setup.CRT[0x01] |= foo & 0xFF;

   foo = (mode->x.blankstart / DCLKdiv);	/* horizontal blank start */
     setup.CRT[0x02] |= foo & 0xFF;

   foo = (mode->x.syncstart / DCLKdiv);		/* horizontal sync start  */
     setup.CRT[0x04] |= foo & 0xFF;

   foo = (mode->x.syncend / DCLKdiv);		/* horizontal sync end    */
     setup.CRT[0x05] |= foo & CR05_HSE_MASK;


   foo = (mode->x.blankend / DCLKdiv); 		/* horizontal blank end   */
     setup.CRT[0x03] |= foo & CR03_HBE_MASK;
     setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
  
   foo = (mode->x.total / DCLKdiv) - 5;		/* horizontal total       */
     setup.CRT[0x00] |= foo & 0xFF;

  /*\
  |*|	setup vertical timing
  \*/
 
   foo = mode->y.width - 1;			/* vertical display end   */
     setup.CRT[0x12] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9  : 0;

   foo = mode->y.blankstart-1;			/* vertical blank start   */
     setup.CRT[0x15] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8  : 0;
     setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9  : 0;
    
   foo = mode->y.syncstart;			/* vertical sync start    */
     setup.CRT[0x10] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9  : 0;

   foo = mode->y.syncend;			/* vertical sync end      */
     setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;
 
   foo = mode->y.blankend + 1;			/* end vertical blank     */
     setup.CRT[0x16] |= foo & 0xFF;
  
   foo = mode->y.total - 2;			/* vertical total         */
     setup.CRT[0x06] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9  : 0;
 
  setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
  setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

  /*\
  |*|	Ok. We'll do the programming now.
  \*/


  /*Program Clock */
  if (mode->clock == -1) INTERNALERROR;
  setup.SEQ [0x0E] = mode->clock>>8;
  setup.SEQ [0x1E] = mode->clock&0xFF;
#ifdef DEBUG
  printk ("Using clock (0x%x,0x%x) which is %d kHz\n",mode->clock>>8,
	  mode->clock&0xFF, (((mode->clock>>8)&0x7F)*28636/
			     (mode->clock&0x3E))>>(mode->clock&1));
#endif

  ENTER_CRITICAL ;
    kgi_ramdac_set_mode ( mode ) ;
    kgi_monitor_set_mode ( mode ) ;
    restore_state( &setup ) ;
  LEAVE_CRITICAL ;

  if ( mode->request.graphtype == GT_TEXT )
    {
      getsetfont ( SETFONT, 0, mode->font0 ) ;
      if ( mode->font1 )
	getsetfont ( SETFONT, 1, mode->font1 ) ;
      else
	getsetfont ( SETFONT, 1, mode->font0 ) ;
    }
  else
    {
	if (chip_version>=CLGD5426_ID)
	{
	  SET_PAGE((chipinfo.ramsize>>16)-1);
	  *((int *)0xAFFF8)=-1;
	  *((int *)0xAFFFC)=-1;
	  SET_PAGE(0);
	  out3GRC(chipinfo.ramsize-8,0x2C);
	  outbGRC(0x0D,0x32);
	};
    }
}

int
kgi_chipset_ioctl ( struct inode *inode, struct file *file, unsigned int cmd,
                  unsigned long arg )
{
  switch(cmd)
    {
      /*
    case CHIPSET_GETINFO:
      _out_buf = &chipinfo; _io_cnt = sizeof(chipinfo);
      return EOK;
    case CHIPSET_SETINFO:
      if (suser())
	{
	  memcpy(&chipinfo, _io_buf, sizeof(chipinfo));
	  return EOK;
	}
      else
	return -EPERM;
	*/
    default:
      return -ECHIPSET ;
    }
}

int drawbox(int x,int y,int w,int h,int color)
{
  int pixw;

  /* Clipping */
  if (x<0 || y<0) return -ENBOGUS; 
  if (x>=current_mode.virtual.x || y>=current_mode.virtual.y) return -EOK;
  if (w<0 || h<0) return -ENBOGUS; 
  if (w==0 || h==0) return -EOK; 
  if ((x+w) > current_mode.virtual.x) w=current_mode.virtual.x-x; 
  if ((y+h) > current_mode.virtual.y) h=current_mode.virtual.y-y;

  /* Draw box */
  SET_PAGE((chipinfo.ramsize>>16)-1);
  *((int *)0xAFFF8)=-1;
  *((int *)0xAFFFC)=-1;
  SET_PAGE(sel_page);

  pixw = (current_mode.graphtype==GT_16BIT) ? 2 : 1;

  outwGRC((w*pixw)-1,0x20);
  outwGRC(h-1,0x22);
  outwGRC(current_mode.virtual.x*pixw,0x24);
  out3GRC((y*(current_mode.virtual.x)+x)*pixw,0x28);
  out3GRC(chipinfo.ramsize-8,0x2C);
  outbGRC(color&0xff,0x01);
  outbGRC((color>>8)&0xff,0x11);
  outbGRC((current_mode.graphtype==GT_8BIT) ? 0xC0 : 0xD0,0x30);
  outbGRC(0x0D,0x32);
  outbGRC(inbGRC(0x31) | 0x02, 0x31);
  while(inbGRC(0x31) & 1);
  /* Can't easily run in background because foreground color has */
  /* to be restored. */
  outbGRC(0x00,0x01);
  return -EOK;
}


int
kgi_graphics_ioctl ( struct inode *inode, struct file *file, unsigned int cmd,
                   unsigned long arg )
{
  struct box *bx = (void *) _io_buf;
  struct kgi_state *gr = file->private_data;
  int err;
 
  if (gr==NULL) {
    panic("%s(%d): Horrible event has happened - kgi_state is NULL!",
	  __FILE__,__LINE__);
    return -ENODRVSUP_ALWAYS_CANT;
  }
  
  if (cmd == GRAPHICS_SETDRAWINFO) {
    gr->drawinfo=*(ggi_drawinfo **)_io_buf;
    return -EOK;
  }

  err=verify_area(VERIFY_READ,gr->drawinfo,sizeof(ggi_drawinfo));
  if (err) return err;
  
  switch(cmd)
    {
    case GRAPHICS_DRAWBOX:
      switch(current_mode.graphtype)
	{
	case GT_8BIT:
	case GT_16BIT:
	  if (chip_version>=CLGD5426_ID)
	    return drawbox(bx->x,bx->y,bx->w,bx->h,
			   get_fs_long(&gr->drawinfo->fg_color));
	default:
	  return -ENODRVSUP_ALWAYS_MMAP;
	};
    case GRAPHICS_FILLSCREEN:
      switch(current_mode.graphtype)
        {
	case GT_8BIT:
	case GT_16BIT:
          if (chip_version>=CLGD5426_ID)
	    return drawbox(0,0,current_mode.virtual.x,current_mode.virtual.y,
			   get_fs_long(&gr->drawinfo->fg_color));
	default:
	  return -ENODRVSUP_ALWAYS_MMAP;
	};
    case GRAPHICS_SET16COL:
      outbGRC((*(int *)_io_buf)& 3,4);       /* Read Map select */
      outbGRC((*(int *)_io_buf)&15,0);       /* Color select */
      return EOK;
    case GRAPHICS_SETPAGE:
      if (graph_buffer)
	{ 
	  sel_page=*((int *)_io_buf);
	  kgi_unmap_all();
	} else 
	  SET_PAGE(sel_page=*((int *)_io_buf));
      return EOK;
    default: return -ENODRVSUP_ALWAYS_MMAP;
    }
  return -ECHIPSET ;
}

#include "mapping.inc"

