/*\
|*| --------------------------------------------------------------------------
|*|  IBM (vga) chipset driver
|*| --------------------------------------------------------------------------
|*|
|*|  Copyright (C) 1996      Jason McMullan, [jmcc@cs.cmu.edu]
|*|
|*|  This program is free software; you can redistribute it and/or modify
|*|  it under the terms of the GNU General Public License as published by
|*|  the Free Software Foundation; either version 2, or (at your option)
|*|  any later version.
|*|
|*|  You should have received a copy of the GNU General Public License
|*|  along with this program; see the file COPYING.  If not, write to
|*|  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|
|*|  This code was derived from the following source of information:
|*|  [1]     Xfree 86 vga driver
\*/

#define	MAINTAINER	"Jason McMullan, [jmcc@cs.cmu.edu]"
#undef	UNDERLINE		/* define to enable underline (attr 01h 09h 81h 89h) */

/*\
|*|  Kernel includes
\*/

extern void request_region(unsigned int from, unsigned int num,
			   const char *name);
extern int check_region(unsigned int from, unsigned int num);
extern void release_region(unsigned int from, unsigned int num);

/*\
|*|     GGI includes
\*/

#define DRIVER_VERSION	"v0.07"

#include <ggi/kgi.h>
#include <ggi/ggi.h>
#include <ggi/graphdev.h>

#include <ggi/IBM/vga.h>
#include "vga_io.inc"

#undef TRACE
#define TRACE(x) x

static struct kgi_chipset chipinfo =
{
    "IBM (or compatable)",	/* manufacturer         */
    "VGA",			/* model                */
    GM_TEXT |          		/* supported modes      */
    GM_4BIT_PLANES |
    GM_8BIT,
    0,
    {900, 540},			/* maximum resolution   */
    0x40000,			/* 256k (quasi-linear) addressable RAM size */
    {25000000, 32000000}	/* DCLK limits            */
};

static struct statetable	/* this table holds the default bits */
 default_state =		/* for text & graphics mode, will be */
{				/* filled in during mode_init        */
    0x23,			/* MISC */
    {
	0x03, 0x00, 0x00, 0x00, 0x00	/* SEQ */
    },
    {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10,		/* CRT */
	0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x1x */
	0xFF,
    },
    {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* GRC */
	0xFF
    },
    {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,		/* ATC */
	0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	0x00, 0x00, 0x0F, 0x00, 0x00
    }
};

static unsigned long text_buffer = 0;
static unsigned long graph_buffer = 0;

static int detect_VGA(void)
{
    fb_text_len = 0x20000;	/* 128k framebuffer  - for mondo text modes? */
    fb_text = kgi_get_physical_mem(&text_buffer, 0x000A0000, 0x000A0000,
			       fb_text_len, fb_text_len, chipinfo.model);

    if (text_buffer == 0) {
	printk("  VGA text memory unavailable \n");
	return -ECHIPSET;
    }
    fb_graph_len = 0x20000;	/* We have only 64K linear graphics */
    fb_graphics = kgi_get_physical_mem(&graph_buffer, 0x000A0000, 0x000A0000,
			     fb_graph_len, fb_graph_len, chipinfo.model);

    if (graph_buffer == 0) {
	printk("  VGA graphics memory unavailable \n");
	return -ECHIPSET;
    }
    return EOK;
}

#define	SETFONT 0
#define	GETFONT 1

static void getsetfont(int get, int no, ggi_font * font)
{
    IS_CRITICAL;
    unsigned char SR2, SR4, GR4, GR5, GR6;
    int i;
    char *fbuf, *data;

    if (font == NULL)
	return;

    fbuf = (char *) (0x000A0000 + (no * 0x2000) + (font->first << 5));
    data = font->data;

    ENTER_CRITICAL;
    SR2 = inbSEQ(0x02);
    SR4 = inbSEQ(0x04);		/* save register state        */
    GR4 = inbGRC(0x04);
    GR5 = inbGRC(0x05);
    GR6 = inbGRC(0x06);

    outbSEQ(SR02_PLANE2, 0x02);	/* font data goes to plane 2  */
    outbSEQ(SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

    outbGRC(2, 0x04);		/* read plane 2               */
    outbGRC(GR5_WRITEMODE0, 0x05);
    outbGRC(GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);
    LEAVE_CRITICAL;
    /* !!! NOTE: no printk()'s are allowed in here unless klogd is active !!!  */

    if (get) {			/* read out font data         */
	if (data)
	    for (i = font->first; i <= font->last; i++) {
		memmove(data, fbuf, font->ysize);
		fbuf += 32;
		data += font->ysize;
	    }
    } else if (data) {		/* font data valid ?          */
	register int cnt = 32 - font->ysize;

	for (i = font->first; i <= font->last; i++) {
	    memmove(fbuf, data, font->ysize);
	    memset(fbuf + font->ysize, 0, cnt);
	    fbuf += 32;
	    data += font->ysize;
	};
    } else			/* clear font data            */
	memset(fbuf, 0, (font->last - font->first + 1) << 5);

    ENTER_CRITICAL;
    outbCRT((inbCRT(0x0A) & ~CR0A_CURSOR_START_MASK)	/* setup cursor       */
	    |(font->cr_start & CR0A_CURSOR_START_MASK), 0x0A);
    outbCRT((inbCRT(0x0B) & ~CR0B_CURSOR_END_MASK)
	    | (font->cr_end & CR0B_CURSOR_END_MASK), 0x0B);
#ifndef UNDERLINE
    font->underline = CR14_UNDERLINE_MASK;
#endif
    outbCRT((inbCRT(0x14) & ~CR14_UNDERLINE_MASK)	/* setup underline    */
	    |(font->underline & CR14_UNDERLINE_MASK), 0x14);

    outbSEQ(SR2, 0x02);
    outbSEQ(SR4, 0x04);		/* restore registers  */
    outbGRC(GR4, 0x04);
    outbGRC(GR5, 0x05);
    outbGRC(GR6, 0x06);
    LEAVE_CRITICAL;
}

/*\
|*|  GGI interface
\*/

int kgi_chipset_init(void)
{
    TRACE(printk("IBM VGA Chipset Driver " DRIVER_VERSION "\n"));

    outbMISC(inb(MISCr));	/* adjust IO base       */

    if (!detect_VGA()) {
	if (!check_region(VGAIO_BASE, VGAIO_REGION))
	    request_region(VGAIO_BASE, VGAIO_REGION, "VGA");
	else
	    printk("%s: warning: failed to allocate IO region.\n", __FILE__);

	return EOK;
    } else
	return -ECHIPSET;
}

void kgi_chipset_done(void)
{
    release_region(VGAIO_BASE, VGAIO_REGION);
    TRACE(printk("VGA Chipset Driver removed\n"));
}

struct kgi_chipset
*kgi_chipset_get_info(void)
{
    return &chipinfo;
}

static void adjust_timing(struct kgi_mode *mode)
{
    register int foo = (mode->request.graphtype == GT_TEXT)
    ? mode->request.textgrid.x : 8;

#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

    ADJUST(width);
    ADJUST(blankstart);
    ADJUST(syncstart);
    ADJUST(syncend);
    ADJUST(blankend);
    if (mode->x.total % foo)
	mode->x.total += foo - (mode->x.total % foo);

    if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
	mode->x.syncend -= foo;
    if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
	mode->x.blankend -= foo;

    if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
	mode->y.syncend--;
    if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
	mode->y.blankend--;

#undef ADJUST
}

static void set_origin(unsigned int offset)
{
    IS_CRITICAL;

    ENTER_CRITICAL;
    outbCRT(offset >> 8, 12);
    outbCRT(offset, 13);
    LEAVE_CRITICAL;
}

static void show_cursor(struct kgi_pointer *cursor, unsigned int offset)
{
    IS_CRITICAL;

    ENTER_CRITICAL;
    cursor->state |= PF_SHOWN;
    outbCRT(offset >> 8, 14);
    outbCRT(offset, 15);
    LEAVE_CRITICAL;
}

static void hide_cursor(struct kgi_pointer *cursor, unsigned int offset)
{
    IS_CRITICAL;

    if (!(cursor->state & PF_SHOWN))
	return;

    ENTER_CRITICAL;
    cursor->state &= ~PF_SHOWN;
    outbCRT(0xFF, 14);
    outbCRT(0xFF, 15);
    LEAVE_CRITICAL;
}

int kgi_chipset_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
    int bpp = 0;

    if (mode->flags & FL_MON_INTERLACE) {
	TRACE(printk("%s: interlaced modes not supported (yet).\n", __FILE__));
	return -ECHIPSET;
    }
    switch (mode->request.graphtype) {
	case GT_8BIT:
            bpp = 8;           /* 1 Bpp */
	    break; 	       
	case GT_4BIT:
	    mode->request.graphtype = GT_4BIT_PLANES;
	case GT_4BIT_PLANES:
	    bpp = 4;           /* 0.5 Bpp */
	    break;	
	case GT_TEXT:
	    break;		/* Not used in textmode */
	default:
	    TRACE(printk("%s: graphtype %i not supported.\n", __FILE__, \
			 mode->request.graphtype));
	    return -ECHIPSET;
    }

    if (mode->lclk.mul != 1) {
	TRACE(printk("%s: invalid lclk multiplier.\n", __FILE__));
	return -ECHIPSET;
    }
    switch (mode->lclk.div) {
	case 1:
	    break;
	default:
	    TRACE(printk("%s: invalid CLK to LCLK divider.\n", __FILE__));
	    return -ECHIPSET;
    }

#define XVIRTUAL (mode->request.virtual.x)
#define YVIRTUAL (mode->request.virtual.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

    switch (cmd) {
	case CMD_PROPOSE:

	    if (mode->request.graphtype == GT_TEXT) {
		mode->magnify.x = mode->magnify.y = 1;
		if ((mode->request.textgrid.x < 8)
		    || (9 < mode->request.textgrid.x)) {
		    TRACE(printk("%s: invalid char width (%i).\n", __FILE__,
				 mode->request.textgrid.x));
		    return -EINVAL;
		}
		if ((2 * (XVIRTUAL / mode->request.textgrid.x)
		* (YVIRTUAL / mode->request.textgrid.y)) > fb_text_len) {
		    TRACE(printk("%s: not enough text memory.\n", __FILE__));
		    return -ENOMEM;
		}
		mode->request.visible.x -=
		    (mode->request.visible.x % mode->request.textgrid.x);
		mode->request.visible.y -=
		    (mode->request.visible.y % mode->request.textgrid.y);
	    } else {		/* mode->request.graphtype == GT_TEXT */
		int ytmp = YVISIBLE;
		int xtmp = XVISIBLE;

		mode->magnify.x = 1;
		mode->magnify.y = 1;

		while (XVISIBLE < 640) {
		    mode->magnify.x++;
		    XVISIBLE += xtmp;
		}

		while (YVISIBLE < 350) {
		    mode->magnify.y++;
		    YVISIBLE += ytmp;
		}

		if (mode->magnify.y > 2)	/* Can't do more zoom than this.. */
		    return -EINVAL;

		if (mode->magnify.x > 2)	/* Can't do more than doublescan... */
		    return -EINVAL;

		if ((mode->request.textgrid.x < 8)
		    || (9 < mode->request.textgrid.x)) {
		    TRACE(printk("%s: invalid char width (%i).\n", __FILE__,
				 mode->request.textgrid.x));
		    return -EINVAL;
		}
		if ((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize) {
		  if ((XVISIBLE * YVISIBLE * bpp / 8) > chipinfo.ramsize) {
		    TRACE(printk("%s: not enough memory to" \
				 " do virtual %ix%i at %i bpp\n", \
				 __FILE__, XVIRTUAL, YVIRTUAL, bpp));
		    return -ENOMEM;
                  } 
                  XVIRTUAL=XVISIBLE;
                  YVIRTUAL=YVISIBLE;
		}
		mode->request.visible.x &= 0x7FFFFFF8;
		mode->request.visible.y &= 0x7FFFFFF8;
	    }

	    mode->x.width = mode->request.visible.x;
	    mode->y.width = mode->request.visible.y;

	    return EOK;

	case CMD_LOWER:

	    if (mode->dclk < chipinfo.dclk.min) {
		TRACE(printk("%s: DCLK too low (%i).\n", __FILE__, mode->dclk));
		return -ECHIPSET;
	    }
	    mode->dclk = (mode->dclk > chipinfo.dclk.max)
		? chipinfo.dclk.max : mode->dclk;

	    adjust_timing(mode);
	    return EOK;

	case CMD_RAISE:

	    if (mode->dclk > chipinfo.dclk.max) {
		TRACE(printk("%s: DCLK too high (%i).\n", \
			     __FILE__, mode->dclk));
		return -ECHIPSET;
	    }
	    mode->dclk = (mode->dclk < chipinfo.dclk.min)
		? chipinfo.dclk.min : mode->dclk;

	    adjust_timing(mode);
	    return EOK;

	case CMD_CHECK:
	    if (mode->request.graphtype == GT_TEXT) {
		mode->set_origin = set_origin;
		mode->cursor_hide = hide_cursor;
		mode->cursor_show = show_cursor;
		mode->flags |= FL_CHIP_COLORTEXT;
		if ((mode->magnify.x != 1) || (mode->magnify.y != 1)	/*driver */
		    ||(mode->request.textgrid.x < 8)	/* hardware */
		    ||(9 < mode->request.textgrid.x)	/* hardware */
		    ||(mode->request.textgrid.y < 1)	/* hardware */
		    ||(mode->request.textgrid.y > 32)	/* hardware */
		    ||((2 * (XVIRTUAL / mode->request.textgrid.x)
		  * (YVIRTUAL / mode->request.textgrid.y)) > fb_text_len)
		    || (mode->dclk > chipinfo.dclk.max)		/* hardware */
		    ||(mode->dclk < chipinfo.dclk.min)) {	/* hardware */
		    TRACE(printk("%s: hw check failed\n", __FILE__));
		    return -ECHIPSET;
		}
	    } else {
		switch (XVIRTUAL) {
		    case 320:
		    case 360:
		    case 640:
		    case 720:
			break;
		    default:
			return -ECHIPSET;
		}

		if (((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize)
		    || (mode->magnify.x > 2) || (mode->magnify.y > 2)
		    || (mode->clk.mul != 1) || (mode->clk.div != 1)
		    || (mode->dclk < chipinfo.dclk.min)
		    || (chipinfo.dclk.max < mode->dclk)
		    || (mode->magnify.x == 2 && mode->request.graphtype != GT_8BIT))
		    return -ECHIPSET;
	    }

#define TM(X) (((unsigned) chipinfo.maxdots.x < mode->x.X)\
		       || ((unsigned) chipinfo.maxdots.y < mode->y.X))
	    if (TM(width) || TM(blankstart) || TM(syncstart)
		|| TM(syncend) || TM(blankend) || TM(total)) {
		TRACE(printk("%s: timing check failed\n", __FILE__));
		return -ECHIPSET;
	    }
#undef TM

	    adjust_timing(mode);
	    return EOK;

	default:
	    INTERNALERROR;
	    return -ECHIPSET;
    }

#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

void kgi_chipset_set_mode(struct kgi_mode *mode)
{
    IS_CRITICAL;
    struct statetable setup = default_state;
    register int foo;
    int DCLKdiv;

    /*\
    |*|        Setup bits per pixel and logical screen width
    \*/

    switch (mode->request.graphtype) {
	case GT_TEXT:
	    foo = 2 * mode->request.virtual.x / mode->request.textgrid.x;
	    foo /= 4;		/* due to the addressing mode used. */
	    break;
	case GT_4BIT_PLANES:
	    foo = mode->request.virtual.x >> 4;
	    break;
	case GT_8BIT:
	    foo = mode->request.virtual.x >> 3;
	    break;
	default:
	    INTERNALERROR;
	    return;
    }

    setup.CRT[CRT_LOGICALWIDTH] |= foo & 0xFF;

    if (mode->request.graphtype == GT_TEXT) {
	setup.CRT[0x17] = CR17_CGA_ADDR_WRAP
	    | CR17_ENABLE_SYNC | CR17_CGA_BANKING
	    | CR17_HGC_BANKING;

	setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;

	setup.GRC[0x05] |= GR5_ODD_EVEN;
	setup.GRC[0x06] |= GR6_MAP_B8_32 | GR6_CHAIN_OE;

	switch ((DCLKdiv = mode->request.textgrid.x)) {
	    case 8:
		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		break;
	    case 9:
		break;
	    default:
		INTERNALERROR;
		return;
	}
	setup.CRT[0x09] |= mode->request.textgrid.y - 1;
	setup.ATC[0x13] = (mode->request.textgrid.x == 9) ? 0x08 : 0x00;
    } else {			/* mode->request.graphtype == GT_TEXT */
	DCLKdiv = 8;

	setup.CRT[0x17] = CR17_CGA_BANKING | CR17_HGC_BANKING
	    | CR17_ENABLE_SYNC | CR17_CGA_ADDR_WRAP;

	for (foo = 0; foo < 16; foo++)
	    setup.ATC[foo] = foo;

	setup.SEQ[0x02] |= SR02_PLANE_MASK;
	setup.SEQ[0x04] |= SR04_256K_ACCESS;

	setup.ATC[0x10] |= AR10_GRAPHICS;

	setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64;

	setup.CRT[0x03] |= CR03_IS_VGA;
	setup.CRT[0x09] |= (mode->magnify.y - 1) & CR09_CHARHEIGHT_MASK;

	switch (mode->request.graphtype) {
	    case GT_4BIT_PLANES:
		setup.GRC[0x01] |= 0x0f;	/* Enable all planes */
		setup.GRC[0x05] |= GR5_WRITEMODE3;
		setup.GRC[0x08] |= 0xff;	/* Enable all bits */
		setup.SEQ[0x04] |= SR04_NO_ODDEVEN | SR04_256K_ACCESS;
		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		setup.CRT[0x17] |= CR17_BYTE_MODE;
		break;
	    case GT_8BIT:
		/* Note that this implies multiply.x==2 */
		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
                setup.SEQ[0x02] |= 0x0f;

                if (mode->request.virtual.x == 320 &&
                    mode->request.virtual.y == 200) {
                   /* Planar */
                   setup.SEQ[0x04] |= SR04_NO_ODDEVEN | SR04_CHAINED;
                   setup.CRT[0x14] |= CR14_DOUBLEWORDMODE;
                } else { 
                   /* Mode X */
		   setup.SEQ[0x04] |= SR04_NO_ODDEVEN;
		   setup.CRT[0x17] |= CR17_BYTE_MODE;
                }

		setup.ATC[0x10] |= AR10_8BIT_MODE;

		setup.GRC[0x05] |= GR5_8BIT_COLOR;
		setup.GRC[0x08] |= 0xff;	/* Enable all bits */

		break;
	    default:
		INTERNALERROR;
		return;
	}
    }

    /*\
    |*|        setup horizontal timing
    \*/

    foo = (mode->x.width / DCLKdiv) - 1;	/* horizontal display end */
    setup.CRT[0x01] |= foo & 0xFF;

    foo = (mode->x.blankstart / DCLKdiv);	/* horizontal blank start */
    setup.CRT[0x02] |= foo & 0xFF;

    foo = (mode->x.syncstart / DCLKdiv);	/* horizontal sync start  */
    setup.CRT[0x04] |= foo & 0xFF;

    foo = (mode->x.syncend / DCLKdiv);	/* horizontal sync end    */
    setup.CRT[0x05] |= foo & CR05_HSE_MASK;


    foo = (mode->x.blankend / DCLKdiv);		/* horizontal blank end   */
    setup.CRT[0x03] |= foo & CR03_HBE_MASK;
    setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;

    foo = (mode->x.total / DCLKdiv) - 5;	/* horizontal total       */
    setup.CRT[0x00] |= foo & 0xFF;

    /*\
    |*|        setup vertical timing
    \*/

    foo = mode->y.width - 1;	/* vertical display end   */
    setup.CRT[0x12] |= foo & 0xFF;
    setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0;
    setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0;

    foo = mode->y.blankstart - 1;	/* vertical blank start   */
    setup.CRT[0x15] |= foo & 0xFF;
    setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8 : 0;
    setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9 : 0;

    foo = mode->y.syncstart;	/* vertical sync start    */
    setup.CRT[0x10] |= foo & 0xFF;
    setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8 : 0;
    setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9 : 0;

    foo = mode->y.syncend;	/* vertical sync end      */
    setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;

    foo = mode->y.blankend + 1;	/* end vertical blank     */
    setup.CRT[0x16] |= foo & 0xFF;

    foo = mode->y.total - 2;	/* vertical total         */
    setup.CRT[0x06] |= foo & 0xFF;
    setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8 : 0;
    setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9 : 0;

    setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
    setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

    /*\
    |*|        Ok. We'll do the programming now.
    \*/

    ENTER_CRITICAL;
    kgi_monitor_set_mode(mode);
    kgi_ramdac_set_mode(mode);
    kgi_clock_set_mode(mode);
    restore_state(&setup);
#ifdef DEBUG_STATE
    dump_state(&setup);
#endif
    LEAVE_CRITICAL;

    if (mode->request.graphtype == GT_TEXT) {
	/* set text fonts   */
	getsetfont(SETFONT, 0, mode->font0);
	if (mode->font1)
	    getsetfont(SETFONT, 1, mode->font1);
	else
	    getsetfont(SETFONT, 1, mode->font0);
    } else {
        /* Do nothing.... */
    }
}

int kgi_chipset_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		      unsigned long arg)
{
    return -ECHIPSET;
}

static int sel_page = 0;

#include "vga_modex_accel.inc"

int kgi_graphics_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		       unsigned long arg)
{
    struct pnt *p = (void *) _io_buf;
    struct hlin *hl = (void *) _io_buf;
    struct fhli *fhl = (void *) _io_buf;
    struct box *bx = (void *) _io_buf;
    struct fbox *fbx = (void *) _io_buf;
    struct kgi_state *gr = file->private_data;
    int err;

    if (gr==NULL) {
      panic("%s(%d): Horrible event has happened - kgi_state is NULL!",
            __FILE__,__LINE__);
      return -ENODRVSUP_ALWAYS_CANT;
    }

    if (cmd == GRAPHICS_SETDRAWINFO) {
      gr->drawinfo=*(ggi_drawinfo **)_io_buf;
      return -EOK;
    }

    err=verify_area(VERIFY_READ,gr->drawinfo,sizeof(ggi_drawinfo));
    if (err)
      return err;


    switch (current_mode.graphtype) {
	case GT_TEXT:
	    return -ENODRVSUP_MODE_MMAP;
	case GT_8BIT:
            if (current_mode.virtual.x==320 &&
                current_mode.virtual.y==200)
              return -ENODRVSUP_MODE_MMAP;

	    switch (cmd) {
	/*
	 * Pixel - ModeX VGA
	 */
                case GRAPHICS_DRAWPIXEL:
                    return drawpixel(p->x,p->y,
                                     get_fs_long(&gr->drawinfo->fg_color));
                case GRAPHICS_PUTPIXEL:
                    return drawpixel(p->x,p->y,p->col);
        /*
         * HLines
         */
                case GRAPHICS_DRAWHLINE:
                    return drawhline(hl->x,hl->y,hl->w,
                                     get_fs_long(&gr->drawinfo->fg_color));
		case GRAPHICS_PUTHLINE:
		    return puthline(fhl->x,fhl->y,fhl->w,fhl->content);

	/*
	 * VLines
	 */
                case GRAPHICS_DRAWVLINE:
                    return drawvline(hl->x,hl->y,hl->w,
                                     get_fs_long(&gr->drawinfo->fg_color));
                case GRAPHICS_PUTVLINE:
                    return putvline(fhl->x,fhl->y,fhl->w,fhl->content);
        /*
         * Box 
         */
                case GRAPHICS_DRAWBOX:
                    return drawbox(bx->x,bx->y,bx->w,bx->h,
                                   get_fs_long(&gr->drawinfo->fg_color));
		case GRAPHICS_PUTBOX:
                    return putbox(fbx->x,fbx->y,fbx->w,fbx->h,fbx->content);

         /*
          * Fill screen
          */
                case GRAPHICS_FILLSCREEN:
                    return fillscreen(get_fs_long(&gr->drawinfo->fg_color));

		default:
		    return -ENODRVSUP_MODE_LOWER;
            }
            return -ENODRVSUP_MODE_LOWER;
	case GT_4BIT_PLANES:
	    switch (cmd) {
		case GRAPHICS_SET16COL:
		    outbGRC((*(int *) _io_buf) & 3, 4);	/* Read Map select */
		    outbGRC((*(int *) _io_buf) & 15, 0);/* Color select */
		    return EOK;
		case GRAPHICS_SETPAGE:
#if 0
		    sel_page = *((int *) _io_buf);/* VGA has only one page. */
		                                  /* ignore silently.       */
		    kgi_unmap_all();
#endif
		    return EOK;
		default:
		    return -ENODRVSUP_MODE_MMAP;
	    }
	    return -ENODRVSUP_MODE_MMAP;
	default:
	    return -ENODRVSUP_ALWAYS_CANT;
    }
    return -ENODRVSUP_ALWAYS_CANT;	/* This driver can't do this. */
}

#include "vga_mapping.inc"
