/*\ $Id: trident.c,v 1.2 1996/07/26 23:48:47 todd Exp $
|*| --------------------------------------------------------------------------
|*|	IBM (vga) chipset driver
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1996	Jason McMullan, [jmcc@cs.cmu.edu]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|
|*|	This code was derived from the following source of information:
|*|	[1]	Xfree 86 vga driver
\*/
/* started from vga.c from Jason M.  Thanks! */
#define	MAINTAINER	"Todd Fries, [todd@miango.com]"
#undef	UNDERLINE	/* define to enable underline (attr 01h 09h 81h 89h) */
#undef  NO_MMAP		/* Until 8 bit works again.... */
/*\
|*|	Kernel includes
\*/

/*
#include <linux/mm.h>
#include <linux/mman.h>
*/

extern void request_region(unsigned int from, unsigned int num,
                           const char *name);
extern int  check_region  (unsigned int from, unsigned int num);
extern void release_region(unsigned int from, unsigned int num);

/*\
|*|     GGI includes
\*/

#define DRIVER_VERSION	"v0.01"

#include <ggi/kgi.h>
#include <ggi/ggi.h>
#include <ggi/graphdev.h>
#include <ggi/mapping.h>

#include <ggi/IBM/vga.h>
#include "vga_io.inc"

static struct kgi_chipset
chipinfo =
{
	"Quadtel",	/* manufacturer		*/
	"Trident xx00",			/* model		*/
	GM_TEXT | GM_1BIT  |	/* supported modes	*/
	GM_4BIT_PLANES | GM_8BIT | GM_16BIT | GM_24BIT,
	0, /* FIX ME!!! max res && RAM should be detected, right? */
	{ 1152, 900 },		/* maximum resolution	*/
	0x100000,		/* 1024k addressable RAM size */
	{ 25000000, 118800000 }	/* DCLK limits		*/
};

static struct statetable	/* this table holds the default bits */
default_state =			/* for text & graphics mode, will be */
{				/* filled in during mode_init        */
  0x23, /* MISC */
  {		
     0x03, 0x00, 0x00, 0x00, 0x00                    	/* SEQ */
  },
  {
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* CRT */
     0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x1x */
     0xFF,
  },
  {
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC */
     0xFF
  },
  {
     0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC */
     0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
     0x00, 0x00, 0x0F, 0x00, 0x00
  }
};

static unsigned long text_buffer = 0;
static unsigned long graph_buffer = 0;

static int detect_VGA(void)
{
  fb_text_len = 0x10000;   /* 64k framebuffer  - for mondo text modes?*/
  fb_text     = kgi_get_physical_mem(&text_buffer,0x000A0000, 0x000A0000,
                                     fb_text_len, fb_text_len, chipinfo.model);

  if (text_buffer == 0) {
    printk("  VGA text memory unavailable \n");
    return -ECHIPSET;
  }

#if 0
  fb_graph_len = 0x20000;
#else
  fb_graph_len = 0x10000;	/* We have only 64K linear graphics buffer - right ? */
#endif
  fb_graphics  = kgi_get_physical_mem(&graph_buffer,0x000A0000, 0x000A0000,
                                    fb_graph_len, fb_graph_len, chipinfo.model);

  if (graph_buffer == 0) {
    printk("  VGA graphics memory unavailable \n");
    return -ECHIPSET;
  }

  return EOK;
}

#define	SETFONT 0
#define	GETFONT 1

static void getsetfont(int get, int no, ggi_font *font)
{
  IS_CRITICAL;
  unsigned char SR2, SR4, GR4, GR5, GR6;
  int i;
  char *fbuf,*data;

  if (font == NULL) return;

  fbuf = (char*) (0x000A0000 + (no * 0x2000) + (font->first << 5));
  data = font->data;

  ENTER_CRITICAL;
    SR2 = inbSEQ(0x02); SR4 = inbSEQ(0x04);	/* save register state        */
    GR4 = inbGRC(0x04); GR5 = inbGRC(0x05); GR6 = inbGRC(0x06);

    outbSEQ(SR02_PLANE2, 0x02);			/* font data goes to plane 2  */
    outbSEQ(SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

    outbGRC(2, 0x04);				/* read plane 2		      */
    outbGRC(GR5_WRITEMODE0, 0x05);
    outbGRC(GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);
  LEAVE_CRITICAL;
  /* !!! NOTE: no printk()'s are allowed in here unless klogd is active !!!  */

  if (get)
    {						/* read out font data         */
      if (data)
        for(i = font->first; i <= font->last; i++)
         {
           memmove(data, fbuf, font->ysize);
           fbuf += 32; data += font->ysize;
         }
    }
   else
    if (data)					/* font data valid ?          */
      {
        register int cnt = 32 - font->ysize;

        for(i = font->first; i <= font->last; i++)
         {
           memmove(fbuf, data, font->ysize);
           memset (fbuf + font->ysize, 0, cnt);
           fbuf += 32; data += font->ysize; 
         };
      }
     else					/* clear font data            */
      memset(fbuf, 0, (font->last - font->first + 1) << 5);

  ENTER_CRITICAL;
    outbCRT(  (inbCRT(0x0A)   & ~CR0A_CURSOR_START_MASK)/* setup cursor       */
            | (font->cr_start &  CR0A_CURSOR_START_MASK), 0x0A);
    outbCRT(  (inbCRT(0x0B)   & ~CR0B_CURSOR_END_MASK)
            | (font->cr_end   &  CR0B_CURSOR_END_MASK), 0x0B);
    #ifndef UNDERLINE
      font->underline = CR14_UNDERLINE_MASK;
    #endif
    outbCRT(  (inbCRT(0x14)   & ~CR14_UNDERLINE_MASK)	/* setup underline    */
            | (font->underline&  CR14_UNDERLINE_MASK), 0x14);

    outbSEQ(SR2, 0x02); outbSEQ(SR4, 0x04);		/* restore registers  */
    outbGRC(GR4, 0x04); outbGRC(GR5, 0x05); outbGRC(GR6, 0x06);
  LEAVE_CRITICAL;
}

/*\
|*|	GGI interface
\*/

int
kgi_chipset_init(void)
{
  TRACE(printk("IBM VGA Chipset Driver "DRIVER_VERSION"\n"));

  outbMISC(inb(MISCr));			/* adjust IO base	*/

  if(! detect_VGA())
    {
      if (! check_region(VGAIO_BASE, VGAIO_REGION))
        request_region(VGAIO_BASE, VGAIO_REGION, "VGA");
       else   
        printk("%s: warning: failed to allocate IO region.\n",__FILE__);

      return EOK;
    }
   else
    return -ECHIPSET;
}

void
kgi_chipset_done(void)
{
  release_region(VGAIO_BASE, VGAIO_REGION);
  TRACE(printk("VGA Chipset Driver removed\n"));
}

struct kgi_chipset 
*kgi_chipset_get_info (void)
{
  return &chipinfo;
}

static void 
adjust_timing(struct kgi_mode *mode)
{
  register int foo = (mode->request.graphtype == GT_TEXT) 
  			   ? mode->request.textgrid.x : 8;

  #define ADJUST(X) mode->x.X -= (mode->x.X % foo)

  ADJUST(width);
  ADJUST(blankstart);
  ADJUST(syncstart);
  ADJUST(syncend);
  ADJUST(blankend);
  if (mode->x.total % foo)
   mode->x.total += foo - (mode->x.total % foo);	

  if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
    mode->x.syncend -= foo;
  if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
    mode->x.blankend -= foo;

  if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
    mode->y.syncend--;
  if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
    mode->y.blankend--;

  #undef ADJUST
}

static void
set_origin(unsigned int offset)
{
  IS_CRITICAL;

  ENTER_CRITICAL;
     outbCRT(offset >> 8, 12);
     outbCRT(offset, 13);
  LEAVE_CRITICAL;
}

static void
show_cursor(struct kgi_pointer *cursor, unsigned int offset)
{
  IS_CRITICAL;

  ENTER_CRITICAL;
    cursor->state |= PF_SHOWN;
    outbCRT(offset >> 8, 14);
    outbCRT(offset, 15);
  LEAVE_CRITICAL;
}

static void
hide_cursor(struct kgi_pointer *cursor, unsigned int offset)
{
  IS_CRITICAL;

  if (!(cursor->state & PF_SHOWN))
    return;

  ENTER_CRITICAL;
    cursor->state &= ~PF_SHOWN;
    outbCRT(0xFF, 14);
    outbCRT(0xFF, 15);
  LEAVE_CRITICAL;
}

int
kgi_chipset_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int bpp=0;

  if (mode->flags & FL_MON_INTERLACE)
    {
      TRACE(printk("%s: interlaced modes not supported (yet).\n",__FILE__));
      return -ECHIPSET; 
    }

  switch(mode->request.graphtype)
   {
     case GT_8BIT:	bpp = 8; break;	/*   1 Bpp	*/
     case GT_4BIT:	TRACE(printk("%s: Fallback GT_4BIT -> GT_4BIT_PLANES\n",__FILE__));
     			mode->request.graphtype=GT_4BIT_PLANES;
     case GT_4BIT_PLANES:
     			bpp = 4; break;	/* 0.5 Bpp	*/
     case GT_1BIT:      bpp = 1; break; /* 1/8 Bpp      */
     case GT_TEXT:	break;          /* Not used in textmode */
     default:
	TRACE(printk("%s: graphtype %i not supported.\n", __FILE__, \
		     mode->request.graphtype));
	return -ECHIPSET;
   }

  if (mode->lclk.mul != 1)
    {
      TRACE(printk("%s: invalid lclk multiplier.\n", __FILE__));
      return -ECHIPSET;
    }

  switch (mode->lclk.div)
   {
     case 1: 
	break;
     default:
	TRACE(printk("%s: invalid CLK to LCLK divider.\n",__FILE__));
	return -ECHIPSET;
   }

  #define XVIRTUAL (mode->request.virtual.x)
  #define YVIRTUAL (mode->request.virtual.y)
  #define XVISIBLE (mode->request.visible.x)
  #define YVISIBLE (mode->request.visible.y)

  switch (cmd)
   {
     case CMD_PROPOSE:

	if (mode->request.graphtype == GT_TEXT)
	  {
	    mode->magnify.x = mode->magnify.y = 1;
	    if (   (mode->request.textgrid.x < 8) 
	        || (9 < mode->request.textgrid.x))
	      {
	        TRACE(printk("%s: invalid char width (%i).\n",__FILE__,
		      mode->request.textgrid.x));
		return -EINVAL;
	      }

	    if ((2 * (XVIRTUAL / mode->request.textgrid.x)
		   * (YVIRTUAL / mode->request.textgrid.y)) > fb_text_len )
	      {
		TRACE(printk("%s: not enough text memory.\n",__FILE__));
		return -ENOMEM;
	      }
	    
	    mode->request.visible.x -= 
		(mode->request.visible.x % mode->request.textgrid.x);
	    mode->request.visible.y -= 
		(mode->request.visible.y % mode->request.textgrid.y);
	  }
         else	/* mode->request.graphtype == GT_TEXT */
	  {
            int ytmp=YVISIBLE;
            int xtmp=XVISIBLE;

            mode->magnify.x = 1;
            mode->magnify.y = 1;

            while (XVISIBLE < 640) {
              mode->magnify.x++;
              XVISIBLE += xtmp;
            }

            while (YVISIBLE < 350) {
              mode->magnify.y++;
              YVISIBLE += ytmp;
            }

            if (mode->magnify.y>15) /* Can't do more zoom than this.. */
              return -EINVAL;

            if (mode->magnify.x>2)  /* Can't do more than doublescan... */
              return -EINVAL;

	    if (   (mode->request.textgrid.x < 8) 
	        || (9 < mode->request.textgrid.x))
	      {
	        TRACE(printk("%s: invalid char width (%i).\n",__FILE__,
		      mode->request.textgrid.x));
		return -EINVAL;
	      }

	    if ((XVIRTUAL* YVIRTUAL* bpp / 8) > chipinfo.ramsize)
	      {
	        TRACE(printk(\
		"%s: not enough memory to do virtual %ix%i at %i bpp",\
		__FILE__, XVIRTUAL, YVIRTUAL, bpp));
	        return -ENOMEM;
	      }

	    mode->request.visible.x &= 0x7FFFFFF8;
	    mode->request.visible.y &= 0x7FFFFFF8;
	}

	mode->x.width = mode->request.visible.x;
	mode->y.width = mode->request.visible.y;

	return EOK;

     case CMD_LOWER:

	if (mode->dclk < chipinfo.dclk.min)
	  {
	    TRACE(printk("%s: DCLK too low (%i).\n",__FILE__, mode->dclk));
	    return -ECHIPSET;
          }

	mode->dclk = (mode->dclk > chipinfo.dclk.max)
	 		  ?  chipinfo.dclk.max : mode->dclk;

	adjust_timing(mode);
	return EOK;

     case CMD_RAISE:

	if (mode->dclk > chipinfo.dclk.max)
	  {
	    TRACE(printk("%s: DCLK too high (%i).\n",__FILE__, mode->dclk));
	    return -ECHIPSET;
          }

	mode->dclk = (mode->dclk < chipinfo.dclk.min)
			?  chipinfo.dclk.min : mode->dclk;

	adjust_timing(mode);
	return EOK;

     case CMD_CHECK:
	if (mode->request.graphtype == GT_TEXT)
	  {
	     mode->set_origin  = set_origin;
	     mode->cursor_hide = hide_cursor;
	     mode->cursor_show = show_cursor;
	     mode->flags |= FL_CHIP_COLORTEXT;
	     if (   (mode->magnify.x != 1) || (mode->magnify.y != 1) /*driver*/
		 || (mode->request.textgrid.x < 8) /* hardware */
		 || (9 < mode->request.textgrid.x) /* hardware */
		 || (mode->request.textgrid.y < 1) /* hardware */
		 || (mode->request.textgrid.y > 32) /* hardware */
		 || ((2 * (XVIRTUAL / mode->request.textgrid.x)
			* (YVIRTUAL / mode->request.textgrid.y)) > fb_text_len)
		 || (mode->dclk > chipinfo.dclk.max) /* hardware */
		 || (mode->dclk < chipinfo.dclk.min)) { /* hardware */
                TRACE(printk("%s: hw check failed\n",__FILE__));
		return -ECHIPSET;
             }
	  }
	 else
	  {
	    switch (XVIRTUAL)	
	     {
	       case 320: case 360: case 640: case 720:
	     	  break;
	       default:
	  	  return -ECHIPSET;
	     }

  	    if (   ((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize)
	        || (mode->magnify.x > 2) || (mode->magnify.y > 2)
	        || (mode->clk.mul != 1)   || (mode->clk.div != 1)
	        || (mode->dclk < chipinfo.dclk.min) 
	        || (chipinfo.dclk.max < mode->dclk)
                || (mode->magnify.x==2 && mode->request.graphtype!=GT_8BIT))
	      return -ECHIPSET;
	  }

	#define TM(X) (((unsigned) chipinfo.maxdots.x < mode->x.X)\
		       || ((unsigned) chipinfo.maxdots.y < mode->y.X))
	if (   TM(width)   || TM(blankstart) || TM(syncstart)
	    || TM(syncend) || TM(blankend)   || TM(total)) {
          TRACE(printk("%s: timing check failed\n",__FILE__));
	  return -ECHIPSET;
        }
 	#undef TM

	adjust_timing(mode);
	return EOK;

     default:
	INTERNALERROR;
	return -ECHIPSET;
   }

  #undef XVIRTUAL 
  #undef YVIRTUAL
  #undef XVISIBLE
  #undef YVISIBLE
}

void
kgi_chipset_set_mode (struct kgi_mode *mode)
{
  IS_CRITICAL;
  struct statetable setup = default_state;
  register int foo;
  int DCLKdiv;

  /*\
  |*|	Setup bits per pixel and logical screen width
  \*/

  switch(mode->request.graphtype)
   {
     case GT_TEXT:
       	foo = 2 * mode->request.virtual.x / mode->request.textgrid.x;
	foo /= 4; /* due to the addressing mode used. */ 
        break;
     case GT_4BIT_PLANES:
	foo = mode->request.virtual.x >> 4;
	break;
     case GT_8BIT:
	foo = mode->request.virtual.x >> 3;
	break;
     default:
	INTERNALERROR;
	return;
   }
   setup.CRT[CRT_LOGICALWIDTH]  |= foo & 0xFF;

   if(mode->request.graphtype == GT_TEXT)
     {
       setup.CRT[0x17] =   CR17_CGA_ADDR_WRAP 
                         | CR17_ENABLE_SYNC   | CR17_CGA_BANKING 
                         | CR17_HGC_BANKING;

       setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;

       setup.GRC[0x05] |= GR5_ODD_EVEN;
       setup.GRC[0x06] |= /*GR6_MAP_A0_128*/
       			  GR6_MAP_B8_32 | GR6_CHAIN_OE;

       switch((DCLKdiv = mode->request.textgrid.x))
	{
	  case 8: setup.SEQ[0x01] |= SR01_8DOT_CHCLK; break;
	  case 9: break;
	  default:
	     INTERNALERROR;
	     return;
	}
       setup.CRT[0x09] |= mode->request.textgrid.y - 1;
       setup.ATC[0x13] = (mode->request.textgrid.x == 9) ? 0x08 : 0x00;
     }
    else /* mode->request.graphtype == GT_TEXT */
     {
       DCLKdiv = 8;

       setup.CRT[0x17] =   CR17_CGA_BANKING | CR17_HGC_BANKING 
                         | CR17_ENABLE_SYNC | CR17_CGA_ADDR_WRAP;
       
       for (foo = 0; foo < 16; foo++) setup.ATC[foo] = foo;

       setup.SEQ[0x02] |= SR02_PLANE_MASK;

       setup.ATC[0x10] |= AR10_GRAPHICS;

       setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64;

       setup.CRT[0x09] |= (mode->magnify.y-1) & CR09_CHARHEIGHT_MASK;

       if (mode->request.graphtype==GT_4BIT_PLANES) {
         setup.GRC[0x01] |= 0x0f;	/* Enable all planes */
         setup.GRC[0x05] |= GR5_WRITEMODE3;
         setup.GRC[0x08] |= 0xff;	/* Enable all bits */
         setup.SEQ[0x04] |= SR04_NO_ODDEVEN | SR04_256K_ACCESS;
         setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
         setup.CRT[0x17] |= CR17_BYTE_MODE;
       }
       if (mode->request.graphtype==GT_8BIT) {
         /* Note that this implies multiply.x==2 */
         setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
         setup.SEQ[0x04] |= SR04_NO_ODDEVEN | SR04_256K_ACCESS;

         setup.ATC[0x10] |= AR10_8BIT_MODE;

	 setup.GRC[0x01] |= 0x0f;		/* Enable all planes */
         setup.GRC[0x05] |= GR5_8BIT_COLOR; 
	 setup.GRC[0x08] |= 0xff;

         setup.CRT[0x14] |= CR14_DOUBLEWORDMODE;
       } 
     }

  /*\
  |*|	setup horizontal timing
  \*/
 
   foo = (mode->x.width / DCLKdiv) - 1;		/* horizontal display end */
     setup.CRT[0x01] |= foo & 0xFF;

   foo = (mode->x.blankstart / DCLKdiv);	/* horizontal blank start */
     setup.CRT[0x02] |= foo & 0xFF;

   foo = (mode->x.syncstart / DCLKdiv);		/* horizontal sync start  */
     setup.CRT[0x04] |= foo & 0xFF;

   foo = (mode->x.syncend / DCLKdiv);		/* horizontal sync end    */
     setup.CRT[0x05] |= foo & CR05_HSE_MASK;


   foo = (mode->x.blankend / DCLKdiv); 		/* horizontal blank end   */
     setup.CRT[0x03] |= foo & CR03_HBE_MASK;
     setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
  
   foo = (mode->x.total / DCLKdiv) - 5;		/* horizontal total       */
     setup.CRT[0x00] |= foo & 0xFF;

  /*\
  |*|	setup vertical timing
  \*/
 
   foo = mode->y.width - 1;			/* vertical display end   */
     setup.CRT[0x12] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9  : 0;

   foo = mode->y.blankstart-1;			/* vertical blank start   */
     setup.CRT[0x15] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8  : 0;
     setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9  : 0;
    
   foo = mode->y.syncstart;			/* vertical sync start    */
     setup.CRT[0x10] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9  : 0;

   foo = mode->y.syncend;			/* vertical sync end      */
     setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;
 
   foo = mode->y.blankend + 1;			/* end vertical blank     */
     setup.CRT[0x16] |= foo & 0xFF;
  
   foo = mode->y.total - 2;			/* vertical total         */
     setup.CRT[0x06] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9  : 0;
 
  setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
  setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

  /*\
  |*|	Ok. We'll do the programming now.
  \*/

  ENTER_CRITICAL;
    kgi_monitor_set_mode(mode);
    kgi_ramdac_set_mode(mode);
    kgi_clock_set_mode(mode);
    restore_state(&setup);
#ifdef DEBUG_STATE
    dump_state(&setup);
#endif
  LEAVE_CRITICAL;

  if (mode->request.graphtype == GT_TEXT) {
    /* set text fonts	*/
    getsetfont(SETFONT, 0, mode->font0);
    if (mode->font1)
      getsetfont(SETFONT, 1, mode->font1);
    else
      getsetfont(SETFONT, 1, mode->font0);
  } else {
    /* Display test pattern */
    int i;
    for (i=0;i<fb_graph_len;i++) {
      ((char *)fb_graphics)[i]=(i&0xff);
    }
  }
}

int
kgi_chipset_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  return -ECHIPSET;
}

static int sel_page=0;

int
kgi_graphics_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		   unsigned long arg)
{
  switch(cmd)
  { case GRAPHICS_DRAWPIXEL:
#if NO_MMAP	/* 1=SLOW */
	{  ggi_coord *h = (ggi_cord *) _io_buf;
	   unsigned offset = h->x+h->y*current_mode.virtual.x;
	   int col=0x12345678;

	    switch(current_mode.graphtype)
	     {
	       case GT_8BIT:	((char*) fb_graphics)[offset] = (char) col;
				return EOK;
			
	       case GT_15BIT:
	       case GT_16BIT:	((short *) fb_graphics)[offset] = (short) col;
				return EOK;

	       case GT_32BIT:	((int *) fb_graphics)[offset] = (int) col;
				return EOK;
	       default:		printk("slowpixel - don't know GT.\n");
	     }
	  } break;
#else
	     return -ENODRVSUP_ALWAYS_MMAP;
#endif
    case GRAPHICS_SET16COL:
	outbGRC((*(int *)_io_buf)& 3,4);       /* Read Map select */
	outbGRC((*(int *)_io_buf)&15,0);       /* Color select */
	return EOK;
    case GRAPHICS_SETPAGE: 
    	sel_page=*((int *)_io_buf);		/* VGA has only one page. */
						/* ignore silently. */
	kgi_unmap_all();
	return EOK;
#ifdef NO_MMAP
    default: return -ENODRVSUP_ALWAYS_LOWER;
#else
    default: return -ENODRVSUP_ALWAYS_MMAP;
#endif
    /* for functions that will probably fallback to something accelerated,
       you should return ALWAYS_LOWER instead */
  }
  return -ENODRVSUP_ALWAYS_CANT;	/* This driver can't do this. */
}

#define MAXMP 100

static struct Map
{ 
	struct mm_struct *mm;
	unsigned long	 mapstart,maplen;
} map[MAXMP],*mp;

static int mapcnt=0;

static int mappedpage=-1;

static inline void
unmap_all(void)
{
#ifndef NO_MMAP
  for(;mapcnt;mapcnt--)
   {
     mp = &map[mapcnt-1];
     fast_unmap_page_range(mp->mm,mp->mapstart,mp->maplen);
     flush_tlb_range(mp->mm,mp->mapstart,mp->mapstart+mp->maplen);
   }
#endif
}

void kgi_unmap_all(void)
{
  unmap_all();
}

/* graph_len really 0x20000 ? not 0x10000 ? */

void kgi_savescreen(void *whereto)
{ kgi_unmap_all();
  if (verify_area (VERIFY_WRITE, (void *) whereto, fb_graph_len)) return;
  memcpy_tofs(whereto,(void *)fb_graphics,fb_graph_len);
}

void kgi_loadscreen(void *wherefrom)
{ kgi_unmap_all();
  if (verify_area (VERIFY_READ, (void *) wherefrom, fb_graph_len)) return;
  memcpy_fromfs((void *)fb_graphics,wherefrom,fb_graph_len);
}

void kgi_linearmap_close(struct vm_area_struct * area)
{	
	/* FIXME ... delete only those mappings that belong */
	/* to the closing task - not too critical, but ... */
	kgi_unmap_all();
}

unsigned long kgi_linearmap_nopage(struct vm_area_struct * area, unsigned long address,
	int no_share)
{
  struct kgi_state *gr = (struct kgi_state *) area->vm_offset;

	/* This is really dumb. !!! Fixme !!! */

	if (gr->graph_num!=fg_console) {	
		/* A backgrounded process accesses graphics mem ? ! */
		schedule();return (0);
        }
	if (address -area->vm_start >= fb_graph_len) 
	{	/* Accessing graphics mem beyond page limit ! */
		TRACE(printk("access at %ld\n",address -area->vm_start));
		current->blocked &= ~(1<<(SIGSEGV-1));
		send_sig(SIGSEGV, current, 1);
        }
#ifdef NO_MMAP
	if (gr->mode.request.graphtype==GT_TEXT)
	{	map[mapcnt].mm=area->vm_mm;
		map[mapcnt].mapstart=area->vm_start;
		map[mapcnt].maplen=0x8000;
		mapcnt++;
		fast_remap_page_range(area->vm_mm,area->vm_start, 
					0xb8000, 0x8000, area->vm_page_prot);
		flush_tlb_range(area->vm_mm,area->vm_start,area->vm_start+0x8000);
	}
	else
	{	map[mapcnt].mm=area->vm_mm;
		map[mapcnt].mapstart=area->vm_start;
		map[mapcnt].maplen=fb_graph_len;
		mapcnt++;
        	fast_remap_page_range(area->vm_mm,area->vm_start,
        				(unsigned long)graph_buffer, 
        				fb_graph_len, area->vm_page_prot);
		flush_tlb_range(area->vm_mm,area->vm_start,area->vm_start+fb_graph_len);
        }
	/* invalidate ok ??? */

        ++current->min_flt;		/* do some statistics */
#endif
        
	return((unsigned long)NULL);
}

void kgi_pagedmap_close(struct vm_area_struct * area)
{	
	/* FIXME ... delete only those mappings that belong */
	/* to the closing task - not too critical, but ... */
	kgi_unmap_all();
}


unsigned long kgi_pagedmap_nopage(struct vm_area_struct * area, unsigned long address,
	int no_share)
{
  struct kgi_state *gr = (struct kgi_state *) area->vm_offset;

	if (gr->graph_num!=fg_console) {	
		/* A backgrounded process accesses graphics mem ? ! */
		schedule();return (0);
        }
	if (address -area->vm_start >= 64*1024) 
	{	/* Accessing graphics mem beyond 64K page limit ! */
		current->blocked &= ~(1<<(SIGSEGV-1));
		send_sig(SIGSEGV, current, 1);
        }
	if (mappedpage!=sel_page)			/* delete ALL mapping for old page */
		kgi_unmap_all();

	if (gr->mode.request.graphtype==GT_TEXT)
	{	map[mapcnt].mm=area->vm_mm;
		map[mapcnt].mapstart=area->vm_start;
		map[mapcnt].maplen=0x8000;
		mapcnt++;
		fast_remap_page_range(area->vm_mm,area->vm_start, 
					0xb8000, 0x8000, area->vm_page_prot);
		flush_tlb_range(area->vm_mm,area->vm_start,area->vm_start+0x8000);
	} else
	{	map[mapcnt].mm=area->vm_mm;
		map[mapcnt].mapstart=area->vm_start;
		map[mapcnt].maplen=0x10000;
		mapcnt++;
        	fast_remap_page_range(area->vm_mm,area->vm_start,
        				(unsigned long)graph_buffer/*+0x10000*sel_page*/, 
        				  0x10000, area->vm_page_prot);
		flush_tlb_range(area->vm_mm,area->vm_start,area->vm_start+0x10000);
        }

/* FIXME ??? Is invalidate() handled with flush_tlb_range ? */
        ++current->min_flt;		/* do some statistics */
        
	return((unsigned long)NULL);
}
