#include <ggi/graphdev.h>

#define MMAP_SIZE    0x200000	/* FIXME ! */
#define MMAP_FR_SIZE 0x010000	/* FIXME ! */

/****************************************************************************/
/* Some useful Macros first - these might even be used for user-level-ex-   */
/* tensions for the GGI-lib that make use of special features in certain    */
/* drivers (before they become integrated into the lib ...)                 */
/****************************************************************************/

/* Check driver capabilities for a given function and remember them : */

#define switcha(request) switch(ioctl_avail[request&0x03ff]&AVM_MASK)

/* ATTENTION : this macro will __RETURN__ (!) if the ioctl succeeds or
   we already have an emulation function! */

#define switchavail(request,param) \
  { int rc;								\
    int rq=request&0x03ff;						\
    if (((rc=-emu_func[rq](request,param))&0xffffff00)==ENODRVSUP)	\
    { ioctl_avail[rq]=rc&0xff; }					\
    else return(rc);							\
  }

/* Set a new emulation function and execute it */

#define setdoemu(request,expire,func,param)			\
  do { int rq=request&0x03ff;					\
       if ((ioctl_avail[request&0x03ff]&AVE_MASK)!=AVE_NOW)	\
       { emu_func   [rq]=func;					\
         if ((ioctl_avail[request&0x03ff]&AVE_MASK)>expire)	\
           ioctl_avail[request&0x03ff]=expire;			\
       }							\
       return func(request,param);				\
     } while(0)

#define SETSIG2(sa, sig, fun, oldsa)	\
	sa.sa_handler = fun;		\
	sa.sa_flags = 0;		\
	sa.sa_mask = 0;			\
        sigaction(sig, &sa, &oldsa);

/********************************************************************************/
/* These are exported for "special" uses only and SHOULD NOT be used !          */
/********************************************************************************/

enum availlist { 	/* Methods */
			AVM_IOCTL =0x00,	/* ioctl is ok - use it */
			AVM_MMAP  =0x01,	/* use mmap */
			AVM_LOWER =0x02,	/* use lower level functions */
			AVM_CANT  =0x03,	/* no support (like setpal in 16M) */
			AVM_MASK  =0x0f,
			/* Expire-times */
			AVE_TEST  =0x00,	/* Test it ... */
			AVE_NOW   =0x00,	/* Retry next time */
			AVE_MODE  =0x10,	/* reset when mode changes */
			AVE_ALWAYS=0x20,	/* Always do it this way   */
			AVE_MASK  =0xf0 };

extern char *ggi_membuffer;
extern char *ggi_pagedbuffer;
extern int ggi_vt_num;
extern int ggi_tty_fd;
extern int gd_hand;

extern int gd_shmid;
extern char *gd_shmem;

extern ggi_drawinfo ggi_drwinfo;

extern enum availlist ioctl_avail[0x400];
extern int (*emu_func[0x400])(int req,void *par);      /* emulation functions */

extern ggi_mode CurrentMode,StartUpMode;

extern void *StartUpState,*BackgroundState;

extern struct sigaction old_handlers[];

enum Events	{ EV_ALL=0,			/* Ignore */
		  EV_TIMER,			/* Timer  */
		  EV_VIDEO,			/* Accel_ready,retrace,... */
		  EV_VT,			/* VT switch */
		  EV_MOUSE,
		  EV_KEY,
		  NUM_EVENTS
		  };

struct Ev_None {};

struct Ev_Timer	{ int ticks; };

enum EvVideo	{ VIDEO_ACCEL_READY,
		  VIDEO_HSYNC,
		  VIDEO_VSYNC };
struct Ev_Video	{ enum	EvVideo	subtype; };

enum EvVt	{ VT_SWITCH_TO,
		  VT_SWITCH_AWAY };
struct Ev_Vt	{ enum	EvVt	subtype;
		  int console; };

enum EvMouse	{ MOUSE_MOVE = 0x01,
		  MOUSE_BT1_PRESS = 0x02,
		  MOUSE_BT2_PRESS = 0x04,
		  MOUSE_BT3_PRESS = 0x08,
		  MOUSE_BT1_RELEASE = 0x10,
		  MOUSE_BT2_RELEASE = 0x20,
		  MOUSE_BT3_RELEASE = 0x40 } ;
struct Ev_Mouse	{ enum	EvMouse	subtype;
		  int xpos,ypos;
		  int buttonstat; };

enum EvKey	{ KEY_PRESS,
		  KEY_RELEASE };
struct Ev_Key	{ enum	EvKey	subtype;
		  int keycode;
		  int ascii;
		  int modifier_state; };

union Evdata {	struct Ev_None  none;
		struct Ev_Timer timer;
		struct Ev_Video video;
		struct Ev_Vt    vt;
		struct Ev_Mouse mouse;
		struct Ev_Key	key;	};

struct ggi_event { enum Events	eventtype;
		   union Evdata evdata; };

extern int lock_count;
extern volatile int release_flag;
extern volatile int acquire_flag;

void releasevt_signal( int n ) ;
void acquirevt_signal( int n ) ;

/********************************************************************************/
/* Exported functions                                                           */
/********************************************************************************/

void discard_avail(void);
void setup_kbd(void);
void reset_kbd(void);
                  
/* Library control */
void ggi_init(void);		/* Call this first */
void ggi_exit(void);		/* and this last - wil NOT reset old mode ! */
void ggi_panic(char *unused);	/* reset old mode,print panic and exit */

/* Mode control */
int ggi_setmode(ggi_mode *tm);			/* Set any desired mode */
int ggi_keepmode(void);			/* keep mode after exit (suser only) */
int ggi_textmode(int cols,int rows,int fontsizex,int fontsizey);	/* set given textmode */
int ggi_graphmode(int xsize,int ysize,int xvirtual,int yvirtual,int type);	/* set given graphics mode */
int ggi_checkgraphmode(int xsize,int ysize,int xvirtual,int yvirtual,int type);	/* check given graphics mode */

int ggi_setviswin(int x,int y);		/* Set the visible window (panning) */

/* Palette and color control */
int ggi_setpalvec(int start,int anz,ggi_color *pp);
int ggi_getpalvec(int start,int anz,ggi_color *pp);

/* Graphics primitives */
int ggi_setcolor(int x);

int ggi_drawpixel(int x,int y);
int ggi_getpixel(int x,int y,void *buffer);
int ggi_putpixel(int x,int y,void *buffer);

int ggi_drawhline(int x,int y,int width);
int ggi_gethline(int x,int y,int width,void *buffer);
int ggi_puthline(int x,int y,int width,void *buffer);

int ggi_drawvline(int x,int y,int width);
int ggi_getvline(int x,int y,int width,void *buffer);
int ggi_putvline(int x,int y,int width,void *buffer);

int ggi_drawbox(int x,int y,int width,int length);
int ggi_getbox(int x,int y,int width,int length,void *buffer);
int ggi_putbox(int x,int y,int width,int length,void *buffer);
int ggi_copybox(int x,int y,int width,int length,int x2,int y2);

int ggi_fillscreen(void);
int ggi_drawcircle(int xcenter,int ycenter,int rad);
int ggi_drawline(int x1,int y1,int x2,int y2);

int ggi_printf(int x,int y,const char *format,...);

int ggi_mon_power(enum powersave state);

/* Administrativa */
int ggi_vt_activate(int number);
int ggi_get_active_vt(void);
int ggi_run_in_bg(int yesno);
int ggi_register_evh(int evtype,void (*EvHand)(struct ggi_event));
int ggi_send_event(struct ggi_event ev);
int ggi_setpage(int page);

/* functions needed by other modules */
void set16col(int col);
int drawpixel4bit_pl(int unused,void *xy);
void takevtcontrol(void);
          
/* Low level driver access */
int ggi_ioctl(int cmd,void *parm);

/* Keyboard */
int ggi_getch(void);
int ggi_kbhit(void);

/* Constants for Textmode */
#define TEXT_FG_BLACK		0x0000
#define TEXT_FG_BLUE		0x0100
#define TEXT_FG_GREEN		0x0200
#define TEXT_FG_CYAN		0x0300
#define TEXT_FG_RED		0x0400
#define TEXT_FG_MAGENTA		0x0500
#define TEXT_FG_BROWN		0x0600
#define TEXT_FG_GREY		0x0700
#define TEXT_FG_DARKGREY	0x0800
#define TEXT_FG_LIGHTBLUE	0x0900
#define TEXT_FG_LIGHTGREEN	0x0a00
#define TEXT_FG_LIGHTCYAN	0x0b00
#define TEXT_FG_LIGHTRED	0x0c00
#define TEXT_FG_LIGHTMAGENTA	0x0d00
#define TEXT_FG_LIGHTBROWN	0x0e00
#define TEXT_FG_YELLOW		0x0e00
#define TEXT_FG_WHITE		0x0f00
#define TEXT_BG_BLACK		0x0000
#define TEXT_BG_BLUE		0x1000
#define TEXT_BG_GREEN		0x2000
#define TEXT_BG_CYAN		0x3000
#define TEXT_BG_RED		0x4000
#define TEXT_BG_MAGENTA		0x5000
#define TEXT_BG_BROWN		0x6000
#define TEXT_BG_GREY		0x7000
#define TEXT_BG_DARKGREY	0x8000
#define TEXT_BG_LIGHTBLUE	0x9000
#define TEXT_BG_LIGHTGREEN	0xa000
#define TEXT_BG_LIGHTCYAN	0xb000
#define TEXT_BG_LIGHTRED	0xc000
#define TEXT_BG_LIGHTMAGENTA	0xd000
#define TEXT_BG_LIGHTBROWN	0xe000
#define TEXT_BG_YELLOW		0xe000
#define TEXT_BG_WHITE		0xf000

