/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _LIB_COMPILE_

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <string.h>
#include <ggi/graphlib.h>

int gd_hand=-1;				 /* Graphics device file handle	*/
char *ggi_membuffer=NULL;		/* mmap-address of graphics space
					   shmem will be here,when switched
					   away,VGA-RAM else */
char *ggi_pagedbuffer=NULL;		/* mmap-address of graphics space
					   shmem will be here,when switched
					   away,VGA-RAM else */


enum availlist ioctl_avail[0x400];	/* map of driver capabilities */
int (*emu_func[0x400])(int req,void *par);	/* emulation functions */

ggi_mode CurrentMode;
void *BackgroundState;

struct sigaction old_handlers[NSIG];

static void int_handler( int v )
{ char buffer[100];
  sprintf(buffer,"Got signal number %d - exiting.\n",v);
  ggi_panic(buffer);
  sigaction(v, &old_handlers[v], NULL);
  raise(v);
}

/*******************************/
/* initialize graphics package */
/*******************************/
void ggi_init(void)
{ struct sigaction siga;
  ggi_drawinfo *drwc;
  struct stat sb;
  char str[40];
  key_t key;
  
  fstat(0,&sb);
  
  sprintf(str,"/dev/graph%d",ggi_vt_num=(sb.st_rdev&0xff));
  key=ftok(str,'G');

  /* Open the graphics device */
  if ((gd_hand=open(str,O_RDWR))==-1) {perror("open");exit(1);}

  discard_avail(); /* Don't know what the driver supports now */

  sprintf(str,"/dev/tty%d",ggi_vt_num);
  if ((ggi_tty_fd=open(str,O_RDWR))==-1) {perror("VT open");exit(1);}

  ggi_membuffer = mmap(
	(caddr_t)0,		/* Map whereever you like   */
	MMAP_SIZE,		/* 2048K - FIXME ! */
	PROT_READ|PROT_WRITE,	/* for reading and writing  */
	MAP_SHARED,		/* shared mapping */
	gd_hand,		/* from the graphics-device */
	MMAP_LINEAR		/* we want the linear buffer*/
    );

  ggi_pagedbuffer = mmap(
	(caddr_t)0,		/* Map whereever you like   */
	MMAP_FR_SIZE,		/* 64K - FIXME ! */
	PROT_READ|PROT_WRITE,	/* for reading and writing  */
	MAP_SHARED,		/* shared mapping */
	gd_hand,		/* from the graphics-device */
	MMAP_FRAME		/* map the paged buffer     */
    );

  drwc=&ggi_drwinfo;
  ggi_ioctl(GRAPHICS_SETDRAWINFO,&drwc);

  if (ggi_membuffer<=0) ggi_membuffer=NULL;	
  /* This is may fail ... we have no membuffer in this case ! */

  /* So now the buffer for the screen contents when switched away.
     We use shared mem, as we can set the mapping address for it. */
     
  gd_shmid = shmget(key, MMAP_SIZE ,IPC_CREAT | 0666);
  if (-1 == gd_shmid) {fprintf(stderr,"Cannot get shared memory.\n");exit(1);}
  
/*  gd_shmem = shmat(gd_shmid, ggi_membuffer, 0);
  if (gd_shmem<=0) {fprintf(stderr,"Cannot attach shared memory.\n");exit(1);}

  if (shmctl(gd_shmid, IPC_RMID, (struct shmid_ds *)0))
    {fprintf(stderr,"Cannot RMID shared memory.\n");exit(1);}*/

  /* Catch signals so we won't end up in graphics mode ... 
     not really needed anymore due to the restore on close behaviour,
     but the kernel will not recognize that the process that issued 
     KD_GRAPHICS has died until the next vt-switch so ... */
  SETSIG2(siga, SIGINT , int_handler , old_handlers[SIGINT ]);
  SETSIG2(siga, SIGFPE , int_handler , old_handlers[SIGFPE ]);
  SETSIG2(siga, SIGILL , int_handler , old_handlers[SIGILL ]);
  SETSIG2(siga, SIGTERM, int_handler , old_handlers[SIGTERM]);
  SETSIG2(siga, SIGQUIT, int_handler , old_handlers[SIGQUIT]);
  SETSIG2(siga, SIGBUS , int_handler , old_handlers[SIGBUS ]);

  takevtcontrol();
  setup_kbd();

}

/*********************************/
/* deinitialize graphics package */
/*********************************/
void ggi_exit(void)
{ close(gd_hand);				/* Close the device */
  reset_kbd();					/* Restore old tty state */
  ioctl(ggi_tty_fd, KDSETMODE, KD_TEXT);	/* Make sure this is Text-Mode */
  if (shmdt(gd_shmem)) {fprintf(stderr,"Cannot detach shared memory.\n");exit(1);}
}

/**************************************/
/* graceful shutdown for fatal errors */
/**************************************/
void ggi_panic(char *why) { ggi_exit();fputs(why,stderr);exit(1); }
