#include <stdio.h>
#include <ggi/graphlib.h>
#include <vga.h>
#include <vgamouse.h>

#define YES 1
#define NO  0

#define OK     0
#define ERROR -1

#define GGI_CHIPSET 0

int __svgalib_console_fd=0;
int __svgalib_tty_fd=0;

static struct modeinfo_s {
   int width;
   int height;
   int mode;
} ModeInfo[] = {
  {  80,  25,GT_TEXT}, /* #define TEXT     0 */
  { 320, 200,GT_4BIT}, /* #define G320x200x16  1 */
  { 640, 200,GT_4BIT}, /* #define G640x200x16  2 */
  { 640, 350,GT_4BIT}, /* #define G640x350x16  3 */
  { 640, 480,GT_4BIT}, /* #define G640x480x16  4 */

  { 320, 200,GT_8BIT}, /* #define G320x200x256 5 */
  { 320, 240,GT_8BIT}, /* #define G320x240x256 6 */
  { 320, 400,GT_8BIT}, /* #define G320x400x256 7 */
  { 360, 480,GT_8BIT}, /* #define G360x480x256 8 */

  { 640, 480,GT_1BIT}, /* #define G640x480x2   9 */
  
  { 640, 480,GT_8BIT}, /* #define G640x480x256 10 */
  { 800, 600,GT_8BIT}, /* #define G800x600x256 11 */
  {1024, 768,GT_8BIT}, /* #define G1024x768x256 12 */
  {1280,1024,GT_8BIT}, /* #define G1280x1024x256 13 */

  { 320, 200,GT_15BIT},/* #define G320x200x32K 14 */
  { 320, 200,GT_16BIT},/* #define G320x200x64K 15 */
  { 320, 200,GT_24BIT},/* #define G320x200x16M 16 */

  { 640, 480,GT_15BIT},/* #define G640x480x32K 17 */
  { 640, 480,GT_16BIT},/* #define G640x480x64K 18 */
  { 640, 480,GT_24BIT},/* #define G640x480x16M 19 */
  
  { 800, 600,GT_15BIT},/* #define G800x600x32K 20 */
  { 800, 600,GT_16BIT},/* #define G800x600x64K 21 */
  { 800, 600,GT_24BIT},/* #define G800x600x16M 22 */

  {1024, 768,GT_15BIT},/* #define G1024x768x32K 23 */
  {1024, 768,GT_16BIT},/* #define G1024x768x64K 24 */
  {1024, 768,GT_24BIT},/* #define G1024x768x16M 25 */

  {1280,1024,GT_15BIT},/* #define G1280x1024x32K 26 */
  {1280,1024,GT_16BIT},/* #define G1280x1024x64K 27 */
  {1280,1024,GT_24BIT},/* #define G1280x1024x16M 28 */

  { 800, 600,GT_4BIT}, /* #define G800x600x16 29 */
  {1024, 768,GT_4BIT}, /* #define G1024x768x16 30 */
  {1280,1024,GT_4BIT}, /* #define G1280x1024x16 31 */

  { 720, 348,GT_1BIT}, /* #define G720x348x2 32	 Hercules emulation mode */

  { 320, 200,GT_32BIT},/* #define G320x200x16M32 33 */
  { 640, 480,GT_32BIT},/* #define G640x480x16M32 34 */
  { 800, 600,GT_32BIT},/* #define G800x600x16M32 35 */
  {1024, 768,GT_32BIT},/* #define G1024x768x16M32 36 */
  {1280,1024,GT_32BIT} /* #define G1280x1024x16M32 37 */
};


#define __LASTMODE 37


static int InGGI=-2;   /* Current active entry of the ModeInfo table */

/**********************
 * Exported variables *
 **********************/

unsigned char *graph_mem=NULL;

/****************
 * Mode setting *
 ****************/
static void _vga_exit(int arg)
{
  exit(arg);
}

int vga_init(void)
{
  int i;
  ggi_init();

  atexit(ggi_exit);
  for (i=1;i<31;i++)
    signal(i,_vga_exit);

  InGGI=-1;
  return OK;
}

int vga_lastmodenumber()
{
  return __LASTMODE;
}

int vga_setmode(int mode)
{
  int x,y,m,err;
  char buff[1024];

  if (mode<0 || mode>__LASTMODE) 
    return ERROR;

  if (InGGI < -1) 
    vga_init();

  x=ModeInfo[mode].width;
  y=ModeInfo[mode].height;
  m=ModeInfo[mode].mode;
  InGGI=mode;

  if (m == GT_TEXT) {
    if (ggi_textmode(x,y,8,16)) {
      sprintf(buff,"Can't set textmode of %dx%d (fontsize=%dx%d)\n",x,y,8,16);
      ggi_panic(buff);
    } 
  } else {
    err=ggi_graphmode(x,y,x,y,m);
    if (err) {
      sprintf(buff,"Can't set graphics mode of %dx%d in %s colors\n",x,y,
                   (m==GT_1BIT) ? "2" :
                   (m==GT_4BIT) ? "16" :
                   (m==GT_8BIT) ? "256" :
                   (m==GT_15BIT) ? "32 thousand" :
                   (m==GT_16BIT) ? "64 thousand" :
                   (m==GT_24BIT) ? "16 million"  :
                   (m==GT_32BIT) ? "16 million (32 bit)" :
                   "an unknown number of");
      ggi_panic(buff);
    }
    graph_mem=ggi_membuffer;
  }

  return OK;
}


int vga_hasmode(int mode)
{
  int x,y,m,err;
  char buff[1024];

  if (mode<0 || mode>__LASTMODE) 
    return NO;

  if (InGGI < -1) 
    vga_init();

  x=ModeInfo[mode].width;
  y=ModeInfo[mode].height;
  m=ModeInfo[mode].mode;

  if (ggi_graphmode(x,y,x,y,m)==0) {
    /* BUGBUG  - should just _check_ for the mode */
    return YES;
  }

  return NO;
}
  
int vga_setflipchar(int c)
{
  return OK;  /* NOTE: This function is obsolete */
}


int vga_clear(void)
{
  ggi_setcolor(0);  /* Is this black? */
  ggi_fillscreen();

  return OK;
}

int vga_flip(void)
{
  /* BUGBUG - should flip to the text VT, 
   * or back again to the graphics VT
   */
  return ERROR;
}


int vga_getxdim(void)
{
  if (InGGI < 0)
    return 0;

  return ModeInfo[InGGI].width;
}


int vga_getydim(void)
{
  if (InGGI < 0)
    return 0;

  return ModeInfo[InGGI].height;
}

int vga_getcolors(void)
{
  if (InGGI < 0)
    return 0;

  switch (ModeInfo[InGGI].mode) {
    GT_TEXT: return 0;
    GT_1BIT: return 2;
    GT_4BIT: return 16;
    GT_8BIT: return 256;
    GT_15BIT: return (1<<15);
    GT_16BIT: return (1<<16);
    GT_24BIT: 
    GT_32BIT: return (1<<24);
    default: return 0;
  }

  return 0;
}

/*********************************
 *     P a l e t t e  I n f o    *
 *********************************/

int vga_setpalette(int index, int red, int green, int blue)
{
  ggi_color clr;

  clr.r=red<<4;
  clr.g=green<<4;
  clr.b=blue<<4;

  return (ggi_setpalvec(index,1,&clr) ? ERROR : OK);
}

int vga_getpalette(int index, int *red, int *green, int *blue)
{
  ggi_color clr;

  if (ggi_getpalvec(index,1,&clr))
    return ERROR;

  (*red)=clr.r;
  (*green)=clr.g;
  (*blue)=clr.b;

  return OK;
}


int vga_setpalvec(int start, int num, int *pal)
{
  return (ggi_setpalvec(start,num,(ggi_color *)pal) ? ERROR : OK);
}

int vga_getpalvec(int start, int num, int *pal)
{
  return (ggi_getpalvec(start,num,(ggi_color *)pal) ? ERROR : OK);
}


/******************
 * Screen control *
 ******************/

int vga_screenoff(void)
{
  /* BUGBUG - this should 'blank' the screen .. */
  return OK;
}
  
int vga_screenon(void)
{ 
  /* BUGBUG - this should 'unblank' the screen .. */
  return OK;
}


/********************
 * Drawing Commands *
 ********************/


int vga_setcolor(int color)
{
  return (ggi_setcolor(color) ? ERROR : OK);
}

int vga_drawpixel(int x, int y)
{
  return (ggi_drawpixel(x,y) ? ERROR : OK);
}

int vga_drawline(int x1, int y1, int x2, int y2)
{
  return (ggi_drawline(x1,y1,x2,y2) ? ERROR : OK);
}

int vga_drawscanline(int line, unsigned char *colors)
{
  int xlen;

  if (InGGI < 0)
    return ERROR;

  xlen=ModeInfo[InGGI].width;

  return (ggi_puthline(line,0,xlen,colors) ? ERROR : OK);
}

int vga_drawscansegment(unsigned char *colors, int x, int y, int length)
{
  return (ggi_puthline(x,y,length,colors) ? ERROR : OK);
}

int vga_getpixel(int x, int y)
{
  int pixel;

  return (ggi_getpixel(x,y,&pixel) ? pixel : ERROR);
}


int vga_getch(void)
{
  return ggi_getch();
}


int vga_dumpregs(void)
{
  /* BUGBUG - We should dump the mode-info here... */
  return OK;
}

/***************************
 * VGA Lib v1.2 extensions *
 ***************************/



vga_modeinfo *vga_getmodeinfo(int mode)
{
  static vga_modeinfo minfo;

  memset(minfo,0,sizeof(vga_modeinfo));

  if (mode<0 ||  mode>__LASTMODE)
    return NULL;
  
  minfo.width=ModeInfo[mode].width;
  minfo.height=ModeInfo[mode].height;
  switch (ModeInfo[mode].mode) {
    case GT_1BIT: minfo.bytesperpixel=0;
                  minfo.linewidth=minfo.width/8;
                  minfo.colors=2;
                  break;
    case GT_4BIT: minfo.bytesperpixel=0;
                  minfo.linewidth=minfo.width/2;
                  minfo.colors=16;
                  break;
    case GT_8BIT: minfo.bytesperpixel=1;
                  minfo.linewidth=minfo.width;
                  minfo.colors=256;
                  break;
    case GT_15BIT:minfo.bytesperpixel=2;
                  minfo.linewidth=minfo.width*2;
                  minfo.colors=(1<<15);
                  break;
    case GT_16BIT:minfo.bytesperpixel=2;
                  minfo.linewidth=minfo.width*2;
                  minfo.colors=(1<<16);
                  break;
    case GT_24BIT:minfo.bytesperpixel=3;
                  minfo.linewidth=minfo.width*3;
                  minfo.colors=(1<<24);
                  break;
    case GT_32BIT:minfo.bytesperpixel=4;
                  minfo.linewidth=minfo.width*4;
                  minfo.colors=(1<<24);
                  break;
         default: break;
  }

  /* BUGBUG - minfo.linewidth_unit  unset 
   *          minfo.linear_arpeture unset
   *          minfo.arpeture_size   unset
   *          minfo.set_arpeture_page(int page) unset
   */
  minfo.chiptype=GGI_CHIPSET;
  minfo.maxlogicalwidth=minfo.linewidth;
  minfo.maxpixels=minfo.width*minfo.height;
  minfo.memory=minfo.linewidth*minfo.height;

  return &minfo;
}

int vga_getdefaultmode(void)
{
  return (InGGI < 0) ? G320x200x256 : InGGI;
}

int vga_getcurrentmode(void)
{
  return (InGGI < 0) ? ERROR : InGGI;
}


int vga_getcurrentchipset(void)
{
  return GGI_CHIPSET;
}

char *vga_getmodename(int mode)
{
  static char buff[1024];
  int m;

  if (InGGI < 0) 
    return NULL;


  m=ModeInfo[mode].mode;

  if (ModeInfo[mode].mode==GT_TEXT)
    sprintf(buff,"TEXT");
  else
    sprintf(buff,"G%dx%dx%s",ModeInfo[mode].width,ModeInfo[mode].height,
                   (m==GT_1BIT) ? "2" :
                   (m==GT_4BIT) ? "16" :
                   (m==GT_8BIT) ? "256" :
                   (m==GT_15BIT) ? "32K" :
                   (m==GT_16BIT) ? "64K" :
                   (m==GT_24BIT) ? "16M"  :
                   (m==GT_32BIT) ? "16M32" : "??");
  
  return buff;
}


int vga_getmodenumber(char *name)
{
  int i,x,y,m;
  char buff[1024];

  if (sscanf(name,"G%dx%dx%s",&x,&y,buff) != 3)
    return ERROR;

  if (strcmp(buff,"2")==0) 
    m=GT_1BIT;
  else if (strcmp(buff,"16")==0)
    m=GT_4BIT;
  else if (strcmp(buff,"256")==0)
    m=GT_8BIT;
  else if (strcmp(buff,"32K")==0)
    m=GT_15BIT;
  else if (strcmp(buff,"64K")==0)
    m=GT_16BIT;
  else if (strcmp(buff,"16M")==0)
    m=GT_24BIT;
  else if (strcmp(buff,"16M32")==0)
    m=GT_32BIT;
  else
    return ERROR;

  for (i=0;i<=__LASTMODE;i++) {
    if (ModeInfo[i].width==x &&
        ModeInfo[i].height==y &&
        ModeInfo[i].mode==m)
      return i;
  }

  return ERROR;
}


/******************
 *  Memory access *
 ******************/

unsigned char *vga_getgraphmem(void)
{
  /* BUGBUG - get pointer to graphics memory??? */
  return NULL;
}


void vga_setpage(int p)
{
  ggi_setpage(p);
}

void vga_setreadpage(int p)
{
  /* BUGBUG - this is badly emulated */
  ggi_setpage(p);
}

void vga_setwritepage(int p)
{
  /* BUGBUG - badly emulated */
  ggi_setpage(p);
}

void vga_setlogicalwidth(int w)
{
  /* BUGBUG - huh??? */
}

void vga_setdisplaystart(int a)
{
  /* BUGBUG - not supported */
}

void vga_waitretrace(void)
{
  /* BUGBUG - not supported */
}

int vga_claimvideomemory(int n)
{
  /* BUGBUG - not supported */
  return ERROR;
}

void vga_disabledriverreport(void)
{
  /* BUGBUG - not supported */
}

int vga_setmodeX(void)
{
  /* BUGBUG - not supported */
  return ERROR;
}


int vga_getmonitortype(void)
{
  /* BUGBUG - badly emulated */
  return 0;
}

/******************
 * Event handling *
 ******************/

int vga_getmousetype(void)
{
  return MOUSE_MOUSESYSTEMS;
}

void vga_setmousesupport(int s)
{
  /* BUGBUG - not supported */
  return;
}


void vga_lockvc(void)
{
  /* BUGBUG - not supported */
}


void vga_unlockvc(void)
{
  /* BUGBUG - not supported */
}


int vga_getkey(void)
{
  return ggi_kbhit();
}

void vga_runinbackground(int s)
{
  /* BUGBUG - not supported */
}


/*****************
 * Writing stuff *
 *****************/

int vga_oktowrite(void)
{
  return YES;
}

void vga_copytoplanar256(unsigned char *virtualp, int pitch,
				  int voffset, int vpitch, int w, int h)
{
  /* BUGBUG - not supported */
  return;
}

void vga_copytoplanar16(unsigned char *virtualp, int pitch,
				   int voffset, int vpitch, int w, int h)
{
  /* BUGBUG - not supported */
  return;
}

void vga_copytoplane(unsigned char *virtualp, int pitch,
		       int voffset, int vpitch, int w, int h, int plane)
{
  /* BUGBUG - not supported */
  return;
}

int vga_setlinearaddressing(void)
{
  /* BUGBUG - not supported */
  return ERROR;
}

void vga_safety_fork(void (*shutdown_routine) (void))
{
  /* BUGBUG - not supported */
  return;
}


/********************
 * Chipset specific *
 ********************/
void vga_setchipset(int c)
{
  /* BUGBUG - not supported */
  return;
}

void vga_setchipsetandfeatures(int c, int par1, int par2)
{
  /* BUGBUG - not supported */
  return;
}

/*********************
 * Text mode support *
 *********************/

void vga_gettextfont(void *font)
{
  /* BUGBUG - not supported */
  return;
}

void vga_puttextfont(void *font)
{
  /* BUGBUG - not supported */
  return;
}

void vga_settextmoderegs(void *regs)
{
  /* BUGBUG - not supported */
  return;
}

void vga_gettextmoderegs(void *regs)
{
  /* BUGBUG - not supported */
  return;
}


/*************************
 * More drawing commands *
 *************************/

int vga_white(void)
{
  if ((InGGI<0)  || ModeInfo[InGGI].mode == GT_TEXT)
    return ERROR;

  switch (ModeInfo[InGGI].mode) {
    case GT_1BIT : return 1;
    case GT_4BIT : return 15;
    case GT_8BIT : return 15;  /* using the default palette... */
    case GT_15BIT: return (1<<15)-1;
    case GT_16BIT: return (1<<16)-1;
    case GT_24BIT: 
    case GT_32BIT: return (1<<24)-1;
    default: return vga_getcolors()-1;
  }

  return ERROR;
}



static const unsigned char ega_red[16] =
     {0, 0, 0, 0, 168, 168, 168, 168, 84, 84, 84, 84, 255, 255, 255, 255};
static const unsigned char ega_green[16] =
     {0, 0, 168, 168, 0, 0, 84, 168, 84, 84, 255, 255, 84, 84, 255, 255};
static const unsigned char ega_blue[16] =
     {0, 168, 0, 168, 0, 168, 0, 168, 84, 255, 84, 255, 84, 255, 84, 255};



int vga_setrgbcolor(int r, int g, int b)
{
  int COL;

  if ((InGGI<0)  || ModeInfo[InGGI].mode == GT_TEXT)
    return ERROR;

  switch (ModeInfo[InGGI].mode) {
    case GT_15BIT:
        COL =
            (b >> 3) +
            ((g >> 3) << 5) +
            ((r >> 3) << 10);
        break;
    case GT_16BIT:
        COL =
            (b >> 3) +
            ((g >> 2) << 5) +
            ((r >> 3) << 11);
        break;
    case GT_24BIT:
    case GT_32BIT:
        COL = b + (g << 8) + (r << 16);
        break;
    default:
        return 0;
    }

    ggi_setcolor(COL);
    return COL;
}

int vga_setegacolor(int c)
{
    if ((InGGI<0)  || ModeInfo[InGGI].mode == GT_TEXT)
      return ERROR;

    if (c < 0)
        c = 0;
    else if (c > 15)
        c = 15;

    switch (ModeInfo[InGGI].mode) {
      case GT_15BIT:
      case GT_16BIT:
      case GT_24BIT:
      case GT_32BIT:
        return vga_setrgbcolor(ega_red[c], ega_green[c], ega_blue[c]);
      default:
        break;
    }

    ggi_setcolor(c);
    return c;
}


void vga_bitblt(int srcaddr, int destaddr, int w, int h, int pitch)
{
  /* BUGBUG - not supported */
  return;
}

void vga_imageblt(void *srcaddr, int destaddr, int w, int h, int pitch)
{
  /* BUGBUG - not supported */
  return;
}

void vga_fillblt(int destaddr, int w, int h, int pitch, int c)
{
  /* BUGBUG - not supported */
  return;
}

void vga_hlinelistblt(int ymin, int n, int *xmin, int *xmax, int pitch, int c)
{
  /* BUGBUG - not supported */
}
  
  
void vga_blitwait(void)
{
  /* BUGBUG - not supported */
}

int vga_ext_set(unsigned what,...)
{
  /* BUGBUG - not supported */
  return ERROR;
}

int vga_accel(unsigned operation,...)
{
  /* BUGBUG - not supported */
  return ERROR;
}

